import numpy as np
from scipy import special as sp
from numba import njit, prange
from numba.typed import List
import math
import os, sys

from .hsh_symmetrization import HSH_symmetrization_matrix, R
from .. import handle as hdl
from ...config import data_type

class odf:
    def __init__(self, n_max, symmetry, orientation_grid, exclude_ghosts=True):
        """Initializes a model to calculate Orientation Distribution Functions from 
        HyperSpherical Harmonics

        Parameters
        ----------
        n_max : int
            highest order of the harmonics expansion
        symmetry : str
            proper point group in notation
        orientation_grid : ndarray, shape(n,3)
            array of orientations in OTP notation
        exclude_ghosts : bool
            if true, skips the half-uneven orders (2,6,10,..)
        """
        print('\tInitializing hyperspherical harmonics model')
        self.odf_mode = 'hsh'
        self.n_max = n_max
        self.symmetry = symmetry
        self.orders = get_orders( symmetry, n_max, info=True, exclude_ghosts=exclude_ghosts )
        self.G_odf = orientation_grid
        # Properties to be initialized when needed:
        self._basis_functions = None
        self._symmetrization_matrix = None
        self._slices_hsh = None
        self._slices_shsh = None

    @property
    def slices_shsh( self ):
        if self._slices_shsh is None:
            # Number sHSHs for each n
            n_sHSHs  =  np.array( [get_NsHSH(self.symmetry,n) for n in self.orders] ) 
            self._slices_shsh = np.array(
                [ [ n_sHSHs[:k].sum(), n_sHSHs[:k+1].sum() ] 
                    for k in range(n_sHSHs.shape[0]) ]) # 'slices' of coefficients as numbers for each n
        return self._slices_shsh

    @property
    def slices_hsh( self ):
        if self._slices_hsh is None:
            n_HSHs = np.array([(n+1)**2 for n in self.orders])
            self._slices_hsh = np.array(
                [ [ n_HSHs[:k].sum(), n_HSHs[:k+1].sum() ] 
                    for k in range(n_HSHs.shape[0]) ]) # 'slices' of coefficients as numbers for each n
        return self._slices_hsh
    
    @property
    def symmetrization_matrix( self ):
        if self._symmetrization_matrix is None:
            # calculate the matrix that transforms HSHs to sHSHs
            # Xsn = HSH_symmetrization_matrix(np.max(orders), symmetry)
            self._symmetrization_matrix = {'0': np.array([[1.]], np.complex128)}
            for n in self.orders[1:]:
                _, csym = cSymmHSH(self.symmetry, n) # get sSHSs and orders
                self._symmetrization_matrix[str(n)] = csym

            # X_new = HSH_symmetrization_matrix(self.n_max, self.symmetry)
            # stop=1
            # self.symmetrization_matrix, self.slices_hsh, self.slices_shsh
        return self._symmetrization_matrix

    @property
    def basis_functions(self):
        if self._basis_functions is None:
            print('\t\tCalculating sHSH basis functions')
            # first calculate the HSHs, then symmetrize, then sum over coefficients
            # n_hsh = self.slices_hsh[-1,1]
            # n_shsh = self.slices_shsh[-1,1]

            n_hsh = np.diff( self.slices_hsh ).flatten()
            n_shsh = np.diff( self.slices_shsh ).flatten()

            self._basis_functions = np.empty((n_shsh.sum(), self.G_odf.shape[0]))
            for i_n, n in enumerate(self.orders):
                lm = np.empty((n_hsh[i_n], 2), dtype=np.int64)
                idx = 0
                for l in range(n+1):
                    for m in range(-l, l+1):
                        lm[idx, 0] = l
                        lm[idx, 1] = m
                        idx += 1
                
                self._basis_functions[self.slices_shsh[i_n,0]:self.slices_shsh[i_n,1]] = compute_basis_functions(
                            self.G_odf, n, lm, 
                            self.symmetrization_matrix[str(n)], 
                        )

            # hsh_basis = Z_all_parallel(self.G_odf, nlm)
            
            # self._basis_functions = np.empty((n_shsh, self.G_odf.shape[0]))
            # for i_n, n_str  in enumerate(self.symmetrization_matrix.keys()):
            #     self._basis_functions[self.slices_shsh[i_n,0]:self.slices_shsh[i_n,1]] = np.real(
            #         self.symmetrization_matrix[n_str] @ hsh_basis[self.slices_hsh[i_n,0]:self.slices_hsh[i_n,1]])
                

        return self._basis_functions

    def get_odf( self, coefficients ):
                
        return coefficients @ self.basis_functions

    def get_odf_batch( self, coefficients_batch ):
        
        odf_batch = odf_batch_parallel(coefficients_batch.astype(data_type), self.basis_functions.astype(data_type))

        # odf_batch = np.empty( (coefficients_batch.shape[0], G_odf.shape[0]) )
        # for c, coefficients in enumerate(coefficients_batch):
        #     odf_batch[c] = coefficients @ shsh_basis

        return odf_batch

    def get_odf_maxima( self, coefficients_batch ):
        i_max = odf_maximum_indices(coefficients_batch.astype(data_type), self.basis_functions.astype(data_type))
        G_max = self.G_odf[i_max]
        return G_max

    def get_odf_centered( self, coefficients ):

        odf = coefficients @ self.basis_functions
        g_mu = self.G_odf[ np.argmax(odf) ]

        c_centered = np.empty_like(coefficients)
        c_centered[0] = coefficients[0]
        for n in range(1,self.orders.size):
            order = self.orders[n]
            R_centering = Rs_n( np.array([-g_mu[0], g_mu[1], g_mu[2]], np.float64),
                                        order, self.symmetrization_matrix[str(order)] )
            c_centered[self.slices_shsh[n,0]:self.slices_shsh[n,1]] = R_centering @ coefficients[self.slices_shsh[n,0]:self.slices_shsh[n,1]]

        return c_centered @ self.basis_functions

    def get_odf_centered_batch( self, coefficients_batch, mean_orientations ):

        C_centered = np.empty_like(coefficients_batch)
        C_centered[:,0] = coefficients_batch[:,0]
        for n in range(1,self.orders.size):
            order = self.orders[n]
            C_centered[:, self.slices_shsh[n,0]:self.slices_shsh[n,1]] = rotate_coefficients_n_parallel(
                np.column_stack( (-mean_orientations[:,0], mean_orientations[:,1], mean_orientations[:,2]) ).astype(np.float64),
                coefficients_batch[:,self.slices_shsh[n,0]:self.slices_shsh[n,1]].astype(np.float64), order, 
                self.symmetrization_matrix[str(order)])


            # R_centering = Rs_n_stack( np.column_stack( (-G_mu[:,0], G_mu[:,1], G_mu[:,2]) ).astype(np.float64),
            #                             order, symmetrization_matrix[str(order)] )
            # for v in range(C_centered.shape[0]):
            #     C_centered[v, slices[n,0]:slices[n,1]] = R_centering[v] @ coefficients_batch[v,slices[n,0]:slices[n,1]]

        odf_batch = odf_batch_parallel(C_centered.astype(data_type), self.basis_functions.astype(data_type))

        # odf_batch = np.empty( (coefficients_batch.shape[0], G_odf.shape[0]) )
        # for c, coefficients in enumerate(C_centered):
        #     odf_batch[c] = coefficients @ shsh_basis

        return odf_batch
    
    def get_odf_std_batch( self, coefficients_batch, mean_orientations ):
        C_centered = np.empty_like(coefficients_batch)
        C_centered[:,0] = coefficients_batch[:,0]
        for n in range(1,self.orders.size):
            order = self.orders[n]
            C_centered[:, self.slices_shsh[n,0]:self.slices_shsh[n,1]] = rotate_coefficients_n_parallel(
                np.column_stack( (-mean_orientations[:,0], mean_orientations[:,1], mean_orientations[:,2]) ).astype(np.float64),
                coefficients_batch[:,self.slices_shsh[n,0]:self.slices_shsh[n,1]].astype(np.float64), order, 
                self.symmetrization_matrix[str(order)])
        
        omega = self.G_odf[:,0]
        std_batch = std_parallel( C_centered.astype(data_type), self.basis_functions.astype(data_type), omega )
        return std_batch

@njit(parallel=True)
def odf_batch_parallel( coefficients_batch, shsh_basis ):
    odf_batch = np.empty((coefficients_batch.shape[0],shsh_basis.shape[1]), coefficients_batch.dtype)
    for c in prange(coefficients_batch.shape[0]):
        odf_batch[c] = coefficients_batch[c] @ shsh_basis
    return odf_batch

@njit(parallel=True)
def odf_maximum_indices( coefficients_batch, shsh_basis ):
    maximum_indices = np.empty((coefficients_batch.shape[0]), np.int32)
    for c in prange(coefficients_batch.shape[0]):
        maximum_indices[c] = np.argmax(coefficients_batch[c] @ shsh_basis)
    return maximum_indices

@njit(parallel=True)
def rotate_coefficients_n_parallel(G_rot, C, order, symmetrization_matrix):
    C_rotated = np.empty_like(C)
    for v in prange(C.shape[0]):
        R_centering = Rs_n( G_rot[v], order, symmetrization_matrix )
        C_rotated[v] = R_centering @ C[v]
    return C_rotated

@njit(parallel=True)
def std_parallel( centered_coefficients_batch, shsh_basis, omega ):
    Std = np.empty( centered_coefficients_batch.shape[0], centered_coefficients_batch.dtype )
    for o in prange(centered_coefficients_batch.shape[0]):
        odf_o = centered_coefficients_batch[c] @ shsh_basis
        Std[o] = np.sqrt( ( odf_o * omega**2 ).sum() / odf_o.sum() )
    return Std

def mason_kernel(c, ns, slices, K):
    """Applies the Mason kernel on the HSH coefficients

    J. K. Mason and O. K. Johnson, ?Convergence of the hyperspherical 
    harmonic expansion for crystallographic texture,? J Appl Crystallogr, 
    vol. 46, no. 6, pp. 1722?1728, Dec. 2013, doi: 10.1107/S0021889813022814.

    Parameters
    ----------
    c : 1D ndarray, float
        set of sHSH coefficients
    ns : 1D ndarray, int
        used orders
    slices : 2D ndarray, int
        indices for finding the orders in the c-array
    K : float
        HSH damping factor to ensure positivity, usually 1 to 2
        
    Return values
    ------------
    c : 1D ndarray, float
        set of modified sHSH coefficients
    """
    for k, Sn in enumerate(slices):
        c[Sn[0]:Sn[1]] = ( 1- ns[k]/(ns[-1]+1) )**K * c[Sn[0]:Sn[1]]

    return c
    
def get_order_weights(c, ns, slices):
    """ Evaluate the contributions from each order to the ODF """
    print('Coefficient weights sum(c^2):')
    weights = []
    for k, Sn in enumerate(slices):
        weights.append( (c[Sn[0]:Sn[1]]**2).sum() )
        print( '\tOrder %d: %.4f' % (ns[k], weights[k]))
    return np.array(weights)

"""
Functions to calculate and rotate hyperspherical harmonics
"""
# Original Matlab code:
# Copyright 2019 Jeremy Mason
#
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.
#
# Rewritten in python by Moritz Frewein 2025

def Z( omega,theta,phi, n,l,m ):
    """Compute the hyperspherical harmonic function Z^n_{l,m}

    This function is used to construct orientation distribution functions
    for axis-angle rotations. theta and phi are the polar and azimutal
    angles, respectively, defining the rotation axis. omega is the rotation
    angle.
    The function uses the scipy package to compute factorials, Gegenbauer
    polynomials and Associated Legendre functions.

    Source: Mason, J.K., and C.A. Schuh. Acta Materialia 56, no. 20 
        (December 2008): 6141?55. https://doi.org/10.1016/j.actamat.2008.08.031.

    Parameters
    ----------
    omega : ndarray, float
        rotation angle, usually ?[0,pi)
    theta : ndarray, float
        polar angle of the rotation axis, usually ?[0,pi)
    phi : ndarray, float
        azimutal angle of the rotation axis, usually ?[0,2pi)
    n, l, m : int/float
        Indices of the hyperspherical harmonic function
        Usually integer or half-integer, with the conditions:
            n >= 0
            l <= n
            -l <= m <= l

    Return values
    ------------
    Z : ndarray, complex
        returns values on the complex plane with |Z| <= 1
    """
    
    #Check indices
    if n < 0:
        print('n smaller than zero, check that!, n= ',n)
    elif l < 0:
        print('l smaller than zero, check that!, l= ',l)
    elif l > n:
        print(' l larger than n, check that!, l= ',l,' n= ',n)
    elif m < l*-1:
        print ('m smaller than -l, check that!, m = ',m,' -l ',l*-1)
    elif m > l:
        print ('m larger than l, check that!, m= ',m,' l= ',l)

    Z = (-1j)**l * 2**(l+0.5) * sp.factorial(l) / (2*np.pi) * \
        np.sqrt( (2*l+1) * sp.factorial(l-m) * (n+1) * sp.factorial(n-l) / \
        ( sp.factorial(l+m) * sp.factorial(n+l+1) ) ) * \
        (np.sin(omega/2))**l * sp.eval_gegenbauer(n-l,l+1,np.cos(omega/2)) * \
        sp.lpmv(m,l,np.cos(theta)) * np.exp(1j*m*phi)
    
    return Z

### Test to check if Z and Z_numba_array are the same
# from textom.src import hsh
# import numpy as np
# n_points = 100
# test_omega = np.pi * np.random.rand(n_points)
# test_theta = np.pi * np.random.rand(n_points)
# test_phi = 2*np.pi * np.random.rand(n_points)
# old_Z = hsh.Z(test_omega,test_theta,test_phi, 4,3,2)
# new_Z = hsh.Z_numba_array(test_omega,test_theta,test_phi, 4,3,2)
# print(f'maximum deviation: {np.max(np.abs(old_Z - new_Z))}') # was about 1e-15
###



# @njit(parallel=True)
# def compute_basis_functions(orientations, nlm, sym_mats, slices_hsh, slices_shsh):
#     n_shsh = slices_shsh.shape[0]
#     n_orient = orientations.shape[0]
#     G = np.empty((slices_shsh[-1,1], n_orient))  # final result

#     for i_n in prange(n_shsh):
#         n1, n2 = slices_hsh[i_n]
#         s1, s2 = slices_shsh[i_n]
#         M = sym_mats[i_n]   # e.g., prepacked array of symmetry matrices

#         # build local hsh_basis
#         nh = n2 - n1
#         hsh_basis_local = np.empty((nh, n_orient), np.complex128)
#         for z in range(nh):
#             n_idx = n1 + z
#             for i in range(n_orient):
#                 hsh_basis_local[z, i] = Z_numba(
#                     orientations[i, 0], orientations[i, 1], orientations[i, 2],
#                     nlm[n_idx, 0], nlm[n_idx, 1], nlm[n_idx, 2]
#                 )

#         # multiply and store
#         G[s1:s2] = np.real(M @ hsh_basis_local)
#     return G

@njit(parallel=True)
def compute_basis_functions(G_odf, n, lm, symm_matrix_n):
    n_shsh, n_hsh = symm_matrix_n.shape
    shsh_basis_n = np.zeros((n_shsh, G_odf.shape[0]), np.float64)
    for z in prange(n_hsh):
        hsh_bf = np.empty((G_odf.shape[0]), np.complex128)
        for i in range(G_odf.shape[0]):
            hsh_bf[i] = Z_numba( G_odf[i,0], G_odf[i,1], G_odf[i,2], 
                                     n, lm[z,0], lm[z,1] )        
        for l in range(n_shsh):
            shsh_basis_n[l] += np.real( symm_matrix_n[l, z] * hsh_bf )
    return shsh_basis_n

# @njit(parallel=True)
# def Z_all_parallel(orientations, nlm):
#     n_hsh = nlm.shape[0]
#     hsh_basis = np.empty((n_hsh, orientations.shape[0]), np.complex128)
#     for z in prange(n_hsh):
#         for i in range(orientations.shape[0]):
#             hsh_basis[z,i] = Z_numba( orientations[i,0], orientations[i,1], orientations[i,2], 
#                                      nlm[z,0], nlm[z,1], nlm[z,2] )
#     return hsh_basis

@njit
def Z_numba(omega, theta, phi, n, l, m):
    """
    Numba-friendly version of Z.
    All inputs are scalars. n,l,m are integers with n>=l>=0 and |m|<=l.
    Returns complex128.
    """
    # sanity (no Python exceptions in njit; keep cheap guards)
    if l < 0 or n < 0 or l > n or abs(m) > l:
        return np.nan + 0j

    # pieces
    half_omega = 0.5 * omega
    s = np.sin(half_omega)
    c_half = np.cos(half_omega)
    ct = np.cos(theta)

    # (-1j)**l
    p_neg_i = pow_neg_i(l)

    # real prefactor before special functions
    # 2**(l+0.5) * fac(l) / (2*pi)
    ln_fac_l = ln_factorial(l)
    ln_pref_real = (l + 0.5) * np.log(2.0) + ln_fac_l - np.log(2.0 * np.pi)
    pref_real = np.exp(ln_pref_real)

    # sqrt( (2l+1) * fac(l-m) * (n+1) * fac(n-l) / ( fac(l+m) * fac(n+l+1) ) )
    ln_num = np.log(2.0 * l + 1.0) \
             + ln_factorial(l - abs(m)) \
             + np.log(n + 1.0) \
             + ln_factorial(n - l)
    ln_den = ln_factorial(l + abs(m)) + ln_factorial(n + l + 1)
    ln_ratio = 0.5 * (ln_num - ln_den)
    sqrt_ratio = np.exp(ln_ratio)

    # (sin(omega/2))**l
    sin_pow = s ** l if l > 0 else 1.0

    # Gegenbauer: C_{n-l}^{(l+1)}(cos(omega/2))
    C = gegenbauer_C(n - l, l + 1.0, c_half)

    # Associated Legendre: P_l^m(cos(theta))
    P = associated_legendre_P(l, m, ct)

    # exp(i m phi)
    e_imphi = np.cos(m * phi) + 1j * np.sin(m * phi)

    # assemble
    out = p_neg_i * (pref_real * sqrt_ratio) * sin_pow * C * P * e_imphi
    return out

# ---------- helpers (nopython-safe) ----------

@njit
def pow_neg_i(l):
    # (-1j)**l cycles every 4
    r = l % 4
    if r == 0:
        return 1.0 + 0.0j
    elif r == 1:
        return 0.0 - 1.0j
    elif r == 2:
        return -1.0 + 0.0j
    else:
        return 0.0 + 1.0j

@njit
def ln_factorial(n):
    # ln(n!) via lgamma
    return math.lgamma(n + 1.0)

@njit
def gegenbauer_C(n, alpha, x):
    # C_n^{(alpha)}(x), n>=0, alpha>0
    if n == 0:
        return 1.0
    if n == 1:
        return 2.0 * alpha * x
    Cnm2 = 1.0
    Cnm1 = 2.0 * alpha * x
    for k in range(1, n):
        # k goes 1..n-1 to produce C_{k+1}
        kp1 = k + 1.0
        num = 2.0 * (k + alpha) * x * Cnm1 - (k + 2.0 * alpha - 1.0) * Cnm2
        Cn = num / kp1
        Cnm2 = Cnm1
        Cnm1 = Cn
    return Cnm1

@njit
def associated_legendre_P(l, m, x):
    """
    Unnormalized associated Legendre P_l^m(x) with Condon-Shortley phase.
    l>=0, |m|<=l, x in [-1,1]
    """
    if m < 0:
        # P_l^{-m} = (-1)^m (l-m)!/(l+m)! P_l^{m}
        mp = -m
        Plmp = associated_legendre_P(l, mp, x)
        sign = -m
        ln_ratio = ln_factorial(l - mp) - ln_factorial(l + mp)
        return ((-1.0)**mp) * np.exp(ln_ratio) * Plmp

    # m >= 0 from here
    # P_m^m(x) = (-1)^m (2m-1)!! (1-x^2)^{m/2}
    # use logs: (2m-1)!! = 2^m * Gamma(m+1/2) / sqrt(pi)
    if m == 0:
        Pmm = 1.0
    else:
        ln_double_fact = m * np.log(2.0) + math.lgamma(m + 0.5) - 0.5 * np.log(np.pi)
        ln_base = 0.5 * m * np.log(max(0.0, 1.0 - x * x))
        Pmm = ((-1.0)**m) * np.exp(ln_double_fact + ln_base)

    if l == m:
        return Pmm

    # P_{m+1}^m(x) = x (2m+1) P_m^m(x)
    Pm1m = x * (2.0 * m + 1.0) * Pmm
    if l == m + 1:
        return Pm1m

    # upward recurrence for l >= m+2
    Plm2 = Pmm
    Plm1 = Pm1m
    for ell in range(m + 2, l + 1):
        num = (2.0 * ell - 1.0) * x * Plm1 - (ell + m - 1.0) * Plm2
        Pl = num / (ell - m)
        Plm2 = Plm1
        Plm1 = Pl
    return Plm1

# -------------------------------------------------------------
@njit
def Rs_n_stack( Gs, n, Xsn ):
    ''' calculates sHSH rotation matrices for all rotation Gs and order n
    '''
    Rs_stack = np.zeros( (Gs.shape[0], Xsn.shape[0], Xsn.shape[0]), np.float64 )
    for g in range(Gs.shape[0]):
        Rs_stack[g] = Rs_n( Gs[g], n, Xsn )
    
    return Rs_stack

@njit
def Rs_n( g, n, Xsn ):
    ''' calculates a single sHSH rotation matrix for a rotation g and order n
    '''
    rot_HSH = R( n, g, np.array([0.,0.,0.]) ) # this is the HSH rotation matrix
    rot_sHSH = np.real( np.conj( Xsn ) @ rot_HSH @ Xsn.T ) # here it's converted to sHSH
    return rot_sHSH


############################################################################################
############################################################################################
### helper functions to access hsh arrays
def get_idx(n,l,m,ns):
    ## input: 
    # n,l,m     desired n,l and m values, e.g. 2,2,0
    # ns        used n's in ascending order, e.g. [0,2,4]
    ## output:
    # idx       index of the coefficient in a 1D array
    
    npos = np.where(ns==n)[0][0]
    ## needed to unroll that one np.sum([2*ll+1 for ll in range(l)])
    idx_temp = 0
    for ll in range(l):
        idx_temp = idx_temp+2*ll+1
    idx = npos + idx_temp + l + m
    return int(idx)

def get_nlm(idx,ns):
    ## input: 
    # idx       index of the coefficient in a 1D array
    # ns        used n's in ascending order, e.g. [0,2,4]
    ## output:
    # n,l,m     corresponding n,l and m values, e.g. 2,2,0

    Ln = [(n+1)**2 for n in ns] # number of unique combinations [l,m] for every n
    Ll = [2*l+1 for l in range(np.max(ns)+1)] # number of m for every l
    ni = 0
    while idx > np.sum(Ln[:ni+1])-1:
        ni += 1
    l = 0
    while idx > np.sum(Ln[:ni]) + np.sum(Ll[:l+1])-1:
        l += 1
    m = idx - np.sum(Ln[:ni]) - np.sum(Ll[:l]) - l
    return ns[ni],l,int(m)

def cSymmHSH(point_group, n):
    """Loads a matrix to transform HSHs to sHSHs

    loads sets of coefficients that make the hyperspherical 
    harmonic expansion invariant a certain crystal symmetry. 
    These can  be interpreted as defining a compact basis
    for an orientation distribution obeying the required symmetry, 
    namely, the symmetrized hyperspherical harmonics (sHSHs).

    This function loads from files created separately by Matlab
    functions in ressources/symmetrizedHSH

    Parameters
    ----------
    proper_point_group : str
        name of the crystal symmetry, taken from a provided table
    n : int
        order of the HSHs

    Return values
    ------------
    nlm : ndarray, int
        HSH indices corresponding to the coefficients
    c : 2D ndarray, complex
        matrix of HSH coefficients for all sHSHs for given n and lattice
        dimensions: 0: HSH orders, 1: sHSHs
    """

    path_csym = hdl.get_file_path('textom',
            os.path.join('ressources','symmetrizedHSH','output',
                         point_group + '_n' + str(n)))
    # filename = 'ressources/symmetrizedHSH/output/' + point_group + '_n' + str(n)
    if os.path.isfile(path_csym):
        data = np.genfromtxt(
            path_csym,
            dtype=np.complex128,
            skip_header=1,
            skip_footer=0,
        )
        
        nlm = data[:,:3].real.astype(int)
        c = data[:,3:].T
        return nlm, c
    
    else:
        print('Symmetrized HSH base file does not exist. Check for typos or generate it via matlab files')
        sys.exit(1)

def get_NsHSH(point_group, n):
    """ Returns how many sHSHs exist for the respective point group and order

    Parameters
    ----------
    point_group : str
        name of the crystal symmetry, taken from a provided table
    n : int
        order of the HSHs
    """
    if np.mod(n,2):
        return 0
    elif n==0:
        return 1
    else:
        path_overview = hdl.get_file_path('textom',
            os.path.join('ressources','symmetrizedHSH','output','overview.txt'))
        with open(path_overview, "r") as file:
                NsHSH_all = eval( file.read() )[point_group]
        return NsHSH_all[int(n/2-1)]


def get_orders( symmetry, n_max = 20, info=True, exclude_ghosts=True ):
    ''' Gives the allowed orders up to n_max
    
    Parameters
    ----------
    n_max : int
        maximum HSH order used
    info : bool
        if True, the list of how many sHSHs exist at each order is printed
    
    Return values
    ------------
    n_allowed : ndarray, int
        orders where sHSHs exist for this point group
    '''
    if info:
        print('\t\torder\tNo of symmetrized HSHs')
    n_allowed = [0]
    for n in range(1,n_max+1):
        Nn = get_NsHSH(symmetry,n)
        if exclude_ghosts and (not n%2) and (n%4): # condition for ghosts
            ghost = f' ({Nn} ghosts)'
            Nn = 0
        else:
            ghost = ''
        if info and Nn > 0:
            print('\t\t%u\t %u%s' % (n, Nn, ghost) )
        if Nn > 0:
            n_allowed.append(n)
    return np.array(n_allowed)
        
def get_order_slices(max_order, symmetry):
    orders = get_orders( symmetry, max_order, info=False )
    n_sHSHs  =  np.array( [get_NsHSH(symmetry,n) for n in orders] ) 

    slices_hsh = np.array(
        [ [ n_sHSHs[:k].sum(), n_sHSHs[:k+1].sum() ] 
            for k in range(n_sHSHs.shape[0]) ]) # 'slices' of coefficients as numbers for each n
    return slices_hsh


import sys, re
from datetime import datetime
import IPython
import textom
import matplotlib.pyplot as plt
from .src import misc as msc
from .src import handle as hdl

def main():
    """Launches TexTOM in iPython mode
    """
    msc.fancy_title()
    plt.ion()
    # Create a dictionary of all public functions and classes from textom
    textom_namespace = {k: v for k, v in vars(textom).items() if not k.startswith("_")}

    log_file_path = "textom_history.log"  # The file to store logs
    with open(log_file_path, "a") as log_file:
        # Redirect stdout and stderr
        sys.stdout = Tee(log_file, sys.stdout)
        sys.stderr = Tee(log_file, sys.stderr)

        # Start IPython with the namespace
        IPython.start_ipython(
            argv=[], 
            # argv=["--logappend", 'textom_history.log'], 
            user_ns=textom_namespace
            )
        
class Tee:
    """A class to write a log file from terminal in and output."""
    ansi_escape = re.compile(r'\x1b\[.*?[@-~]')  # Regex to match ANSI escape sequences

    def __init__(self, log_file, stream):
        self.log_file = log_file  # File object for logging
        self.stream = stream      # Original stream (e.g., sys.stdout or sys.stderr)

    def write(self, message):
        # Always write the raw message to the terminal
        self.stream.write(message)

        # Strip ANSI escape sequences for the log file
        stripped_message = self.ansi_escape.sub('', message)

        # Add a timestamp and write to the log file if not ignored
        if not self._should_ignore(stripped_message):
            timestamped_message = self._add_timestamp(stripped_message)
            self.log_file.write(timestamped_message)
            self.log_file.flush()  # Ensure it's written immediately

    def flush(self):
        """Flush both the log file and the original stream."""
        self.log_file.flush()
        self.stream.flush()

    def isatty(self):
        """Delegate the check to the original stream."""
        return self.stream.isatty()

    def fileno(self):
        """Delegate the fileno method to the original stream."""
        return self.stream.fileno()

    def _should_ignore(self, message):
        """Determine if a message should be ignored for logging."""
        ignore_patterns = [
            r'^In \[',       # IPython input prompts
            r'^Out\[\d+\]',  # IPython output prompts
        ]
        for pattern in ignore_patterns:
            if re.search(pattern, message):
                return True
        # also ignore carriage return updates
        if '\r' in message and '\n' not in message:
            return True
        return False

    def _add_timestamp(self, message):
        """Add a timestamp to the message for logging."""
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        # Add timestamp only to non-empty lines
        return f"{timestamp} {message}" if message.strip() else message

def config():
    """Opens the config file to modify
    """
    config_path = hdl.get_file_path('textom','config.py')
    hdl.open_with_editor(config_path, confirm=False)

def documentation():
    """Opens the TexTOM documentation in a pdf viewer
    """
    doc_path = hdl.get_file_path('textom','documentation/textom_documentation.pdf')
    hdl.open_pdf(doc_path)


import os
import fabio
import h5py
import pyFAI
from time import time
import numpy as np
import argparse, sys
import importlib

from textom.src.integration.pyfai_plugins import flexible_integrator, update_filelist
from textom.src.misc import import_module_from_path

def parallel_launcher( k_task, sample_dir ):
    """starts the pyfai integration and is optimized to be called on parallel CPUs

    Parameters
    ----------
    k_task : int
        task index
    sample_dir : str
        textom sample base directory
    """
    # import integration parameters
    intpar_path_sample = os.path.join(sample_dir,'integration_parameters.py')
    par = import_module_from_path('integration_parameters', intpar_path_sample)
    
    fid_in, filtered_datasets = update_filelist(sample_dir,par)

    flat=None
    if isinstance( par.flatfield_correction, str ):
        flat = fabio.open(par.flatfield_correction).data

    mask = fabio.open(par.mask_path).data
    ai = pyFAI.load(par.poni_path)

    t0 = time()
    cnt = 0
    n_tot = len(filtered_datasets)
    for l in range ( k_task, n_tot, par.n_tasks ):
        try:
            flexible_integrator(sample_dir, fid_in, filtered_datasets[l], par, ai, flat, mask)
        except:
            pass
        
        cnt += 1
        print('\tTask %d: %d/%d done, av. time per scan: %.2f s' % (
            k_task, l+1, n_tot, (time()-t0)/cnt))
    fid_in.close()

# def import_module_from_path(module_name, file_path):
#     spec = importlib.util.spec_from_file_location(module_name, file_path)
#     module = importlib.util.module_from_spec(spec)
#     sys.modules[module_name] = module
#     spec.loader.exec_module(module)
#     return module

def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("-k", "--k_task", type=int, default=0)
    parser.add_argument("-d", "--dir_out_full", type=str, default=0)
    # argcomplete.autocomplete(parser)
    args = parser.parse_args()

    parallel_launcher(args.k_task,args.dir_out_full)

if __name__ == "__main__":
    main(sys.argv[1:])


import numpy as np
from numba import njit, prange

from .. import numba_plugins as nb

@njit(parallel=True)
def order_parameters_parallel( vectors, odfs ): #odf_par, basis_functions, V_fz ):

    Order_parameters = np.empty( odfs.shape[0], vectors.dtype )
    std_director = np.empty( odfs.shape[0], vectors.dtype )
    Directors = np.empty( (odfs.shape[0],3), vectors.dtype )

    for k in prange( odfs.shape[0] ):
        weights = odfs[k] #clipped_odf( V_fz, basis_functions, odf_par[k])
        weights[weights<0] = 0.
        weights /= weights.sum()
        Order_parameters[k], Directors[k], std_director[k] = weighted_order_parameter(vectors,weights)

    return Order_parameters, Directors, std_director

# @njit()
# def clipped_odf( V_fz, basis_functions, coefficients ):
#     """ Computes a orientation distribution function from a set of sHSH coefficients

#     """
#     # renormalize to c0 = 1    
#     c1plus = coefficients[1:]/coefficients[0]

#     ## produces an ODF from HSH-coefficients
#     odf = 1/V_fz + c1plus @ basis_functions

#     odf[odf<0] = 0 # clip negative values
#     odf = odf / odf.sum() # renormalize

#     return odf

@njit
def weighted_order_parameter(vectors, weights):
    Q = np.zeros((3, 3), vectors.dtype)
    for i in range(len(vectors)):
        n = vectors[i]
        norm = np.sqrt(np.dot(n, n))
        n /= norm
        w = weights[i]
        for j in range(3):
            for k in range(3):
                Q[j, k] += w * (1.5 * n[j] * n[k] - 0.5 * (j == k))

    eigvals, eigvecs = jacobi_eigenvalue_3x3(Q)
    max_idx = np.argmax(eigvals)
    S = eigvals[max_idx]
    director = eigvecs[:, max_idx]
    # theta = np.arccos(nb.nb_clip_array(np.dot(vectors, director), -1, 1)) # is there a case where this is important?
    theta = np.arccos(nb.nb_clip_array(np.abs(np.dot(vectors, director)), 0, 1))
    std_director = np.average(theta, weights=weights) * 180/np.pi
    return S, director, std_director

@njit
def outer_product(a, b):
    result = np.zeros((3, 3), a.dtype)
    for i in range(3):
        for j in range(3):
            result[i, j] = a[i] * b[j]
    return result

@njit
def normalize(v):
    norm = np.sqrt(np.sum(v**2))
    return v / norm

@njit
def jacobi_eigenvalue_3x3(A, max_iterations=100, eps=1e-10):
    V = np.eye(3, dtype=A.dtype)
    for iteration in range(max_iterations):
        # Find the largest off-diagonal element
        a01 = abs(A[0, 1])
        a02 = abs(A[0, 2])
        a12 = abs(A[1, 2])

        if a01 >= a02 and a01 >= a12:
            p, q = 0, 1
        elif a02 >= a12:
            p, q = 0, 2
        else:
            p, q = 1, 2

        if abs(A[p, q]) < eps:
            break

        theta = 0.5 * (A[q, q] - A[p, p]) / A[p, q]
        t = np.sign(theta) / (abs(theta) + np.sqrt(1 + theta**2))
        c = 1 / np.sqrt(1 + t**2)
        s = t * c

        # Rotation
        for i in range(3):
            if i != p and i != q:
                aip = A[i, p]
                aiq = A[i, q]
                A[i, p] = A[p, i] = c * aip - s * aiq
                A[i, q] = A[q, i] = c * aiq + s * aip

        app = A[p, p]
        aqq = A[q, q]
        apq = A[p, q]

        A[p, p] = c**2 * app - 2 * s * c * apq + s**2 * aqq
        A[q, q] = s**2 * app + 2 * s * c * apq + c**2 * aqq
        A[p, q] = A[q, p] = 0.0

        for i in range(3):
            vip = V[i, p]
            viq = V[i, q]
            V[i, p] = c * vip - s * viq
            V[i, q] = c * viq + s * vip

    # Eigenvalues are the diagonal of A
    eigenvalues = np.array([A[0, 0], A[1, 1], A[2, 2]], A.dtype)
    return eigenvalues, V


from setuptools import setup, find_packages
import os

def get_version():
    version_file = os.path.join(os.path.dirname(__file__), 'textom', 'version.py')
    with open(version_file) as f:
        # This will read the version from the version.py file
        exec(f.read())
    return locals()['__version__']

setup(
    name="TexTOM",          
    version=get_version(),       
    author="Moritz Frewein, Marc Allain, Tilman Gruenewald",
    author_email="textom@fresnel.fr",
    description="A program for texture simulations.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://gitlab.fresnel.fr/textom/textom.git", 
    packages=find_packages(), 
    include_package_data=True,
    package_data={
        "textom": [
        "ressources/symmetrizedHSH/output/*",
        "ressources/*.txt",
        "input/*.py",
        "documentation/functions/*.tex",
        "documentation/introduction.tex",
        "documentation/textom_documentation.tex",
        "documentation/textom_documentation.pdf",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9, <=3.13',
    install_requires=["mumott >= 2.0, <=2.2", 
                      "pyfai >=2023, <=2025.1", 
                      "orix >= 0.13.0, <=0.13.3", 
                      "psutil==7.0",
                    #   "ase==3.25.0",
                      "threadpoolctl==3.6",
                      "ipython", 
    ],
    entry_points={
        "console_scripts": [
            "textom=textom.entries:main",
            "textom_config=textom.entries:config",
            "textom_documentation=textom.entries:documentation",
        ]
    },
)


# textom/__init__.py
import pkgutil
import importlib
import psutil
import os
from threadpoolctl import threadpool_limits

from .src.misc import get_affinity_info
from .config import n_threads
from .version import __version__

# parallelisation settings
aff_info = get_affinity_info() # this is to avoid conflicts with parallelisation via taskset
if not aff_info['restricted']:
    # set everything to single threaded
    threadpool_limits(limits=1)
    os.environ["MKL_NUM_THREADS"] = "1" 
    os.environ["NUMEXPR_NUM_THREADS"] = "1" 
    os.environ["OMP_NUM_THREADS"] = "1"
    os.environ["OPENBLAS_NUM_THREADS"] = "1"
    from numba import set_num_threads
    n_threads = min(psutil.cpu_count(), n_threads) # make sure the number of threads doesn't exceed the possible ones
    set_num_threads(n_threads)

# import textom modules
__all__ = []
for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    if module_name == 'textom':  # Restrict to the textom module
        module = importlib.import_module(f"{__name__}.{module_name}")
        for name in dir(module):
            if not name.startswith("_"):  # Skip private attributes
                globals()[name] = getattr(module, name)
                __all__.append(name)

# Set package metadata
__author__ = "Moritz Frewein, Moritz Stammer, Marc Allain, Tilman Gruenewald"
__email__ = "textom@fresnel.fr"

import sys # this is to display symbols correctly
sys.stdout.reconfigure(encoding='utf-8')


import numpy as np
import h5py
import os
import matplotlib.pyplot as plt
from pathlib import Path

def list_newest_files(folder, num_files=5):
    """
    Lists the newest files in a folder based on modification time.

    Parameters:
    folder (str): Path to the folder.
    num_files (int): Number of newest files to list (default: 5).

    Returns:
    list: A list of the newest filenames, sorted by modification time (newest first).
    """
    folder_path = Path(folder)

    # Ensure the folder exists
    if not folder_path.is_dir():
        raise ValueError(f"The specified folder does not exist: {folder}")

    # Get all files in the folder with their modification times
    files = [
        (file, file.stat().st_mtime) for file in folder_path.iterdir() if file.is_file()
    ]

    # Sort files by modification time (newest first)
    files_sorted = sorted(files, key=lambda x: x[1], reverse=True)

    # Extract the filenames of the newest files
    newest_files = [file[0].name for file in files_sorted[:num_files]]

    return newest_files

path = '/dataSATA/data_synchrotron/esrf_id13_2024_05_ch7039_enthesisII/20240501/PROCESSED_DATA/INTEG/cement_pmma/analysis/fits/'
MDL, nmax = [],[]
for file in list_newest_files(path, 5):
    with h5py.File(os.path.join(path,file), 'r') as hf:
        MDL.append(hf['MDL'][()])

    i0 = file.find('nmax')
    i1 = file.find('_', i0)
    nmax.append( int( file[i0+4 : i1] ) )

plt.plot(nmax, MDL)
plt.xlabel('HSH order')
plt.ylabel('MDL')
plt.show()


import os
import glob
import numpy as np
from time import time
import sys, shutil
import importlib as imp # for reading input files
from numba import njit, prange
import h5py
import hdf5plugin

# domestic
from .. import handle as hdl
from . import model_projection as prj
from . import model_crystal as cry
from ..odf import hsh
from ..odf import gridbased as grd
from . import rotation as rot
from . import symmetries as sym
from .. import mask as msk
from ..numba_plugins import integrate_c
from ..misc import import_module_from_path
from ...config import data_type, odf_resolution
from ...version import __version__
# from . import model_difflets_odftt as dif

class model_textom:
    """ A class to that contains the theoretical description of Texture tomography.

    """
    def __init__(self, sample_dir, classic=False, single=False, q_det=False, chi_det=False, light=False, no_Isc=False, override_odf_mode=False ):
        """Initializes the texture tomography model class

        Calculates arrays that do not need to be updated when fitting/plotting

        Parameters
        ----------
        startFileName : str
            path to the input file
        single : bool
            set to True if you just want to calculate a single image
            disables the whole tomography part
        
        Attributes created
        ------------
        see returned variables in input file
        """

        print("Initializing model")
        self.title = os.path.basename(sample_dir)
        self.path_analysis = os.path.join(sample_dir,'analysis')
        self.light = light
        self.no_Isc = no_Isc

        ################ Projectors
        if single: # initializes parameters if only a single image is produced
            self._init_single()
        else:
            if os.path.isfile(os.path.join(sample_dir,'analysis','projectors.h5')): # loads a tomo model if it exists
                t0 = time()
                with h5py.File( os.path.join(sample_dir,'analysis','projectors.h5'), 'r' ) as hf:
                    self.Omega = hf['Omega'][()].astype(data_type)
                    self.Kappa = hf['Kappa'][()].astype(data_type)
                    self.Gs = hf['Gs'][()].astype(data_type)
                    self.tomogram = hf['tomogram'][()].astype(data_type)
                    try:
                        self.ty = hf['translations_y'][()].astype(data_type)
                        self.tz = hf['translations_z'][()].astype(data_type)
                    except:
                        self.ty = hf['ty'][()].astype(data_type)
                        self.tz = hf['tz'][()].astype(data_type)
                    self.shift_y = hf['shift_y'][()].astype(data_type)
                    self.shiftz = hf['shift_z'][()].astype(data_type)
                    self.fov = hf['fov'][()]
                    self.x_p = hf['x_p'][()].astype(data_type)
                    self.nVox = hf['nVox'][()]
                    self.mask_voxels = hf['mask_voxels'][()]
                    if not light:
                        self.Beams = hf['Beams'][()].astype(data_type)
                        self.iBeams = hf['iBeams'][()]
                print("\tLoaded projectors from analysis/projectors.h5 (%2.2f s)" % ( (time()-t0)) )
            
            else:
                self.Kappa, self.Omega, self.shift_y, self.shiftz, self.tomogram, self.sinogram = hdl.load_shifts_mumott(
                        os.path.join(sample_dir,'analysis/alignment_result.h5') )

                print('\tImporting geometry')
                geo = import_module_from_path('geometry', os.path.join(sample_dir,'analysis/geometry.py'))

                self.nVox, self.fov, self.Gs, ty, tz, self.x_p = prj.setup_geometry( 
                    geo, self.tomogram, self.Omega, self.Kappa )

                # mask the sample
                if not os.path.isfile(os.path.join(self.path_analysis,'voxelmask.txt')) and np.any(self.sinogram):
                    self.mask_voxels = msk.mask_voxels(self.Kappa, self.Omega, ty, tz, self.shift_y, self.shiftz,
                                                        self.x_p, self.tomogram, self.sinogram)
                        
                    # save voxel mask to file
                    with open(os.path.join(self.path_analysis,'voxelmask.txt'), 'w') as fid:
                        for iv in self.mask_voxels:
                            fid.write(f'{iv}\n')      
                    print('\t\tSaved voxelmask to analysis/voxelmask.txt')  
                elif not np.any(self.sinogram):
                    self.mask_voxels = np.arange(self.tomogram.size)
                else:
                    self.mask_voxels = np.genfromtxt(
                        os.path.join(self.path_analysis,'voxelmask.txt'),
                        np.int32 )

                # calculate beam intensities and save them
                self.Beams, self.iBeams = prj.get_projectors(geo, self.Gs, self.nVox, self.x_p,
                                                                self.mask_voxels, ty, tz, self.shift_y, self.shiftz)

                print('\tSaving results to sample path analysis/projectors.h5')
                with h5py.File( os.path.join(sample_dir,'analysis','projectors.h5'), 'w' ) as hf:
                    hf.create_dataset('Beams', data=self.Beams, compression=hdf5plugin.Blosc(cname='zstd', clevel=5, shuffle=hdf5plugin.Blosc.SHUFFLE))
                    hf.create_dataset('iBeams', data=self.iBeams,compression=hdf5plugin.Blosc(cname='zstd', clevel=5, shuffle=hdf5plugin.Blosc.SHUFFLE))                    
                    hf.create_dataset('Omega', data=self.Omega)
                    hf.create_dataset('Kappa', data=self.Kappa) 
                    hf.create_dataset('Gs', data=self.Gs) 
                    hf.create_dataset('tomogram', data=self.tomogram, compression=hdf5plugin.Blosc(cname='zstd', clevel=5, shuffle=hdf5plugin.Blosc.SHUFFLE))
                    hf.create_dataset('translations_y', data=ty) 
                    hf.create_dataset('translations_z', data=tz) 
                    hf.create_dataset('shift_y', data=self.shift_y) 
                    hf.create_dataset('shift_z', data=self.shiftz) 
                    hf.create_dataset('fov', data=self.fov) 
                    hf.create_dataset('x_p', data=self.x_p)#, compression="lzf") 
                    hf.create_dataset('nVox', data=self.nVox) 
                    hf.create_dataset('mask_voxels', data=self.mask_voxels) 
                    hf.create_dataset('textom_version', data=__version__)
        
        ################# Diffractlets
        crystal_path = os.path.join(sample_dir,'analysis','crystal.py')
        self.cr = import_module_from_path('crystal', crystal_path)
        if override_odf_mode:
            self.odf_mode = override_odf_mode
        else:
            self.odf_mode = self.cr.odf_mode
        if self.odf_mode=='hsh':
            self.odf_module = hsh
            self.odf_parameter = self.cr.hsh_max_order
        elif self.odf_mode=='grid':
            print('grid mode currently not functional')
            self.odf_module = grd
            self.odf_parameter = self.cr.grid_resolution
            return 0
            
        ## Load or calculate stuff
        difflets_path = os.path.join(sample_dir,'analysis',f'difflets_{self.odf_mode}_{self.odf_parameter}.h5')
        if os.path.isfile(difflets_path):
            with h5py.File(difflets_path, 'r') as hf:
                self.difflets = hf['difflets'][()].astype(data_type)
                self.powder_pattern = hf['powder_pattern'][()].astype(data_type)
                self.Qq_det = hf['q_values'][()].astype(data_type)
                self.Chi_det = hf['chi_values'][()].astype(data_type)
                self.hkl = hf['hkl'][()]
                self.detector_reciprocal_coordinates = hf['detector_reciprocal_coordinates'][()].astype(data_type)
                self.symmetry = hf['symmetry'][()].decode('utf-8')
                try:
                    self.structure_factors = hf['structure_factors'][()]
                except:
                    self.structure_factors =  np.ones(self.hkl.shape[0])
                #
                # here check if odf_parameters are the same (?)
                #
            print('\tLoaded diffractlets')

            # Initialize ODF object
            Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=odf_resolution*np.pi/180, 
                                    sampling='cubochoric', symmetry=self.symmetry )
            self.odf = self.odf_module.odf(self.odf_parameter, sym.get_ppg_notation(self.symmetry), Gc)

        else:
            # print('\tImporting geometry')
            geo = import_module_from_path('geometry', os.path.join(sample_dir,'analysis/geometry.py'))

            # get detector coordinates
            if np.any(q_det) and np.any(chi_det):
                # self.q_det = q_det
                self.chi_det = chi_det * np.pi/180
            else:
                # get detector coordinates from integrated data
                int_data_file = glob.glob(os.path.join(sample_dir,'data_integrated','*2d.h5'))[0]
                with h5py.File(os.path.join(sample_dir,'data_integrated',int_data_file),'r') as hf:
                    # self.q_det = hf['radial_units'][()]
                    self.chi_det = hf['azimuthal_units'][()] * np.pi/180
                # # adapt q to desired range
                # self.q_det = self.q_det[np.logical_and(
                #     self.q_det > self.cr.q_range[0],
                #     self.q_det < self.cr.q_range[1],
                # )]

            # print('calculating diffractlets')                
            sample_rotations = rot.QfromOTP(self.Gs)
            self.Qq_det, self.Chi_det, self.detector_reciprocal_coordinates,\
                self.hkl, self.structure_factors, self.difflets, self.powder_pattern, self.symmetry, self.odf = cry.get_diffractlets(
                            self.cr, self.chi_det, geo, 
                            sample_rotations,
                            cutoff_structure_factor=self.cr.cutoff_structure_factor, max_hkl=self.cr.max_hkl,
                            odf_mode=self.odf_mode,
                            hsh_max_order = self.odf_parameter, grid_resolution=self.odf_parameter, # might want to make these a single arg
                            )
            # cry.plot_diffractlet(Qq_det, Chi_det, hkl, difflets[0], q_bins=np.linspace(*cr.q_range),
            #                      cmap='plasma', sym_cmap=False, logscale=True
            #                      )
            # cry.plot_diffractlet(Qq_det, Chi_det, hkl, difflets[1], q_bins=np.linspace(*cr.q_range),
            #                      cmap='plasma', sym_cmap=False, logscale=True
            #                      )
            with h5py.File(difflets_path,'w') as hf:
                hf.create_dataset('difflets', data=self.difflets)
                hf.create_dataset('powder_pattern', data=self.powder_pattern)
                hf.create_dataset('cif_file', data=self.cr.cifPath)
                hf.create_dataset('Energy_keV', data=self.cr.E_keV)
                hf.create_dataset('q_values', data=self.Qq_det)
                hf.create_dataset('hkl', data=self.hkl)
                hf.create_dataset('structure_factors', data=self.structure_factors)
                hf.create_dataset('chi_values', data=self.Chi_det)
                hf.create_dataset('detector_reciprocal_coordinates', data=self.detector_reciprocal_coordinates)
                hf.create_dataset('sample_rotations', data=self.Gs)
                hf.create_dataset('symmetry', data=self.symmetry)
                hf.create_dataset('odf_mode', data=self.odf_mode )
                hf.create_dataset('odf_parameter', data=self.odf_parameter )
                hf.create_dataset('textom_version', data=__version__)

        self.detShape = self.Qq_det.shape

    """
    Initialisation functions
    """          
    def _init_single(self):
        """ Defines experiment-related parameters for simulating only single image """
        self.fov = np.array([1,1])
        self.nVox = np.array([1,1,1])
        self.Omega = np.array([0])
        self.Kappa = np.array([0])
        self.Gs = np.array([[0,0,0]])
        self.x_p = np.array([[0,0,0]])
        self.ty = np.array([0])
        self.tz = np.array([0])
        self.Beams = np.array([[[1,0]]])
        self.iBeams = np.array([[[0,2**32-1]]])
        self.mask_voxels = np.array([True])

    def imageFromC(self, c):
        """ Computes a diffraction image from a set of sHSH coefficients

        Parameters
        ----------
        c : 1D ndarray, float
            set of sHSH coefficients
        
        Return values
        ------------
        image : ndarray, float
            array of scattering intensity for each point on the detector
        """
        image = _imageFromC( c, self.difflets )
        return image
    
    def projection( self, g, C ):
        """Simulates diffraction images from a 2D scan over the sample

        For a certain sample rotation calculates sHSH coefficients in 
        each voxel, then integrates over the beam intensity for each
        translation, calculates the diffraction patterns and saves them
        into the model object.

        Parameters
        ----------
        g : int
            index of the sample rotation, defined in self.samplerotations
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Attributes modified
        ------------
        images : 3D ndarray, float
            array of resulting scattering intensity for each point on the 
            detector, for each rotation and translation
            dim: 0: rotation, 1: translation, 2: detector points
        """
        # print('\tcalculating images for projection %d' % g )
        t0 = time()
        dlt_shp = self.difflets.shape
        images = _projection(
            g, C, 
            self.Beams, self.iBeams, self.difflets[g].reshape((dlt_shp[1], dlt_shp[2]*dlt_shp[3])),
            ).reshape((self.Beams.shape[1], dlt_shp[2], dlt_shp[3]))
        # print("\t\tfinished in %.3f s" % (time()-t0))
        return images
    
    def odfFromC( self, c, recenter = False ):
        """Computes an ODF from sHSHs

        Sums over pre-calculated symmetrized spherical harmonic functions
        and weights by the given coefficients. Also adds a isotropic part
        weighted by another input parameter

        Parameters
        ----------
        c : 1D ndarray, float
            set of sHSH coefficients
        info: bool
            if True prints information about the ODF
            
        Return values
        ------------
        odf : 1D ndarray, float
            array of probabilities for each orientation self.Gc
        """
        # if not hasattr(self.odf, 'basis_functions'):
        #     Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=odf_resolution*np.pi/180, 
        #                             sampling='cubochoric', symmetry=self.symmetry )
        #     self.odf._basis_functions(Gc)

        # calculate ODF
        if recenter:
            odf = self.odf.get_odf_centered( c )
        else:
            odf = self.odf.get_odf( c )
        return self.odf.G_odf, odf
    '''
    g_pref = mod.g_ML_sample( fit.C[mod.mask_voxels], truncate=truncate_expansion )
    results['g_pref'] = mod.insert_sparse_tomogram(g_pref)
    print('\tExtracting stds')
    stds = mod.std_sample( fit.C[mod.mask_voxels], g_pref )
    results['std']=mod.insert_sparse_tomogram(stds)
    # export vtk-file for paraview
    a_pref, b_pref, c_pref = mod.abc_pref_sample( g_pref )
    '''

    def preferred_orientations( self, C ):

        # if not hasattr(self.odf, 'basis_functions'):
        #     Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=odf_resolution*np.pi/180, 
        #                                  sampling='cubochoric', symmetry=self.symmetry )
        #     self.odf._basis_functions(Gc)

        # odfs = self.odf.get_odf_batch( C )
        # ig = np.argmax(odfs, axis=1) # index of the maximum of the odf
        # G_max = self.odf.G_odf[ig]

        G_max = self.odf.get_odf_maxima( C )
        return G_max

    def std_sample( self, C, G_mu ):

        # if not hasattr(self.odf, 'basis_functions'):
        #     Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=odf_resolution*np.pi/180, 
        #                                  sampling='cubochoric', symmetry=self.symmetry )
        #     self.odf._basis_functions(Gc)

        # odfs_centered = self.odf.get_odf_centered_batch( C, G_mu )
        # Std = std_parallel( odfs_centered, self.odf.G_odf[:,0] )

        Std = self.odf.get_odf_std_batch( C, G_mu )

        return Std

    def abc_pref_sample( self, g_pref, lattice_vectors ):
        """ Calculates the crystal axis orientations from OTP orientations

        Should be generalized to all crystal systems using the actual axes

        Parameters
        ----------
        g_pref : 2D ndarray, float
            dim: 0: voxel index, 1: [ome,tta,phi]

        returns: 
        ------------
        a_pref, b_pref, c_pref : 2D ndarray, float
            crystal axis unit vectors
            dim: 0: voxel index, 1: [x,y,z]
        """
        lattice_vectors_norm = lattice_vectors/np.linalg.norm(lattice_vectors,axis=1)
        a_pref, b_pref, c_pref = _abc_pref_sample( g_pref, lattice_vectors_norm.astype(data_type) )
        return a_pref, b_pref, c_pref

    def directions_from_orientations( self, direction_0=(0,0,1) ):
        Gc, dV, V_fz = rot.sample_fundamental_zone( dchi=odf_resolution*np.pi/180, 
                                         sampling='cubochoric', symmetry=self.symmetry )
        return _directions_from_orientations(Gc, np.array(direction_0,data_type)), Gc
    
    def insert_sparse_tomogram( self, values ):
        """ Makes a 3D tomogram out of sparse data
        """
        try:
            tomogram = np.empty( (self.nVox.prod(),values.shape[1]), data_type )
            tomogram[:] = np.nan
            tomogram[self.mask_voxels] = values
            tomogram = tomogram.reshape((*self.nVox, values.shape[1]))
        except:
            tomogram = np.empty( self.nVox.prod(), data_type )
            tomogram[:] = np.nan
            tomogram[self.mask_voxels] = values
            tomogram = tomogram.reshape(self.nVox)
        return tomogram

    def odfFromC_old( self, c, K = None, clip = False, info = False ):
        """Computes an ODF from sHSHs

        Sums over pre-calculated symmetrized spherical harmonic functions
        and weights by the given coefficients. Also adds a isotropic part
        weighted by another input parameter

        Parameters
        ----------
        c : 1D ndarray, float
            set of sHSH coefficients
        K : float or None for not applying
            HSH damping factor to ensure positivity, usually between 1 and 2
        clip: bool
            if True, negative values get clipped and ODF renormalized
        info: bool
            if True prints information about the ODF
            
        Return values
        ------------
        odf : 1D ndarray, float
            array of probabilities for each orientation self.Gc
        """
        if K == 'auto':
            K, neg = 1., 1.
            c_in=c.copy()
            while neg>0.01:
                c_in = self.apply_kernel( c, K, info=False )
                odf = _odfFromC( self.sHSHs, c )
                neg = ( odf[odf<0] @ self.dV[odf<0] ) 
                K += 0.1
        elif isinstance(K,(int, float)): # apply the Mason kernel on coefficients
            c_in = self.apply_kernel( c, K )
        else:
            c_in = c.copy()
        
        # calculate ODF
        odf = _odfFromC( self.sHSHs, c_in )

        if clip:
            neg_percent = - odf[odf<0].sum() / np.abs(odf).sum()
            odf[odf<0] = 0 # clip negative values
            odf = odf / ( odf @ self.dV ) # renormalize

        if info:
            self.get_c_weights(c_in)
            print('Made ODF')
            print('\tMaximum: %.3e' % odf.max() )
            print('\tMinimum: %.3e' % odf.min() )


            # estimate mean value from distribution truncated at lowest order
            if c.size > 1:
                ic_max = self.Sn[1,1]
                i_mu = np.argmax(
                    _odfFromC( self.sHSHs[:ic_max-1], c[:ic_max]))
                mu = self.Gc[i_mu]
                std = self.std_sample(
                    np.array([c_in]), np.array([mu])
                )[0]
            print('\tMean orientation: ome %.2f, tta %.2f, phi %.2f' % (
                mu[0],mu[1],mu[2]))
            print('\tStandard deviation [%s]: %.2f' % (chr(176), std*180/np.pi))
            
            if np.any(odf<0):
                odf_pos = odf.copy()
                odf_pos[odf_pos<0]=0
                tmp = odf_pos @ self.dV - 1
                rel_neg = tmp / (1 + 2*tmp)
                print('\t%.2f percent of the distribution is negative' % (rel_neg*100) )

        return odf

"""
Class-external functions compiled in numba and called from wrappers inside the class
"""
@njit()
def _imageFromC( c, difflets ):
    """Computes a diffraction image from custom parameters
    Statically typed to enable jitting
    Weights pre-calculated sHSH diffraction patterns with coefficients
    to make a custom diffraction pattern

    Parameters
    ----------
    c : 1D ndarray, float
        set of sHSH coefficients
    difflets : 2D ndarray, float
        diffractlets to be summed
        dim 0: sHSH index, 1: detector point index
    
    Return values
    ------------
    image : 1D ndarray, float
        simulated diffraction pattern
    """
    image = c @ difflets # sum over sHSH-images without isotropic part
    return image

@njit()
def _odfFromC( sHSHs, c ):
    """ Computes a diffraction image from a set of sHSH coefficients

    Parameters
    ----------
    c : 1D ndarray, float
        set of sHSH coefficients
    
    Return values
    ------------
    odf : 1D ndarray, float
        probability for each orientation
    """
    # renormalize to c0 = 1    
    c1plus = c[1:]/c[0]

    ## produces an ODF from HSH-coefficients
    odf = 1 + c1plus @ sHSHs

    return odf.astype(data_type)

@njit(parallel=True)
def std_parallel( Odf, omega ):
    Std = np.empty( Odf.shape[0], Odf.dtype )
    for o in prange(Odf.shape[0]):
        Std[o] = np.sqrt( ( Odf[o] * omega**2 ).sum() / Odf[o].sum() )
    return Std

@njit(parallel=True)
def _directions_from_orientations( orientations, direction_0 ):
    """ Calculates the crystal axis orientations from OTP orientations

    Should be generalized to all crystal systems using the actual axes

    Parameters
    ----------
    g_pref : 2D ndarray, float
        dim: 0: voxel index, 1: [ome,tta,phi]

    returns: 
    ------------
    a_pref, b_pref, c_pref : 2D ndarray, float
        crystal axis unit vectors
        dim: 0: voxel index, 1: [x,y,z]
    """
    Nvox = orientations.shape[0]
    directions = np.empty((Nvox,3), data_type)

    for k in prange(Nvox):
        # get preferred orientation for all axes
        R_pref  = rot.MatrixfromOTP( orientations[k,0], orientations[k,1], orientations[k,2])
        directions[k] = direction_0 @ R_pref 
    return directions

@njit(parallel=True)
def _abc_pref_sample( g_pref, lattice_vectors ):
    """ Calculates the crystal axis directions from OTP orientations

    Parameters
    ----------
    g_pref : 2D ndarray, float
        dim: 0: voxel index, 1: [ome,tta,phi]

    returns: 
    ------------
    a_pref, b_pref, c_pref : 2D ndarray, float
        crystal axis unit vectors
        dim: 0: voxel index, 1: [x,y,z]
    """
    g_pref, lattice_vectors = g_pref.astype(data_type), lattice_vectors.astype(data_type)
    Nvox = g_pref.shape[0]
    a_pref = np.empty((Nvox,3), data_type)
    b_pref = np.empty((Nvox,3), data_type)
    c_pref = np.empty((Nvox,3), data_type)

    for k in prange(Nvox):
        # get preferred orientation for all axes
        R_pref  = rot.MatrixfromOTP( g_pref[k,0], g_pref[k,1], g_pref[k,2])
        a_pref[k] = lattice_vectors[0] @ R_pref 
        b_pref[k] = lattice_vectors[1] @ R_pref 
        c_pref[k] = lattice_vectors[2] @ R_pref 
    return a_pref, b_pref, c_pref

@njit(parallel=True)
def _projection(
        g, C, # parameters
        Beams, iBeams, difflets_rot, # pre-calculated quantities
        ):
    """Simulates diffraction images from a 2D scan over the sample

    For a certain sample rotation calculates sHSH coefficients in 
    each voxel, then integrates over the beam intensity for each
    translation and makes diffraction patterns from diffractlets

    Parameters
    ----------
    g : int
        rotation index
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for rotation g
        dim: 0: basefunction indices, 1: detector points

    returns: 
    ------------
    images : 3D ndarray, float
        modified simulated diffraction patterns
        dim: 0: rotation, 1: translation, 2: detector points
    """
    images = np.empty((Beams.shape[1], difflets_rot.shape[1]), data_type)
    for t in prange( Beams.shape[1] ): # scan over the sample
        iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C)
        # calculate the image
        images[t,:] = c_proj @ difflets_rot 
    return images




# textom/src/__init__.py
import pkgutil
import importlib

from . import analysis, data_treatment, integration, model, odf, optimization

__all__ = ['analysis', 'data_treatment', 'integration', 'model', 'odf', 'optimization']

# # Dynamically import all submodules and expose their contents
# __all__ = []
# for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
#     module = importlib.import_module(f"{__name__}.{module_name}")
#     for name in dir(module):
#         if not name.startswith("_"):  # Skip private attributes
#             globals()[name] = getattr(module, name)
#             __all__.append(name)




import inspect
import re

import textom.textom

def generate_latex_doc(module, output_file="library_functions.tex", filter_keyword=""):
    """
    Generates a LaTeX document listing all functions of a library with their docstrings.

    Parameters:
    module (module): The library module to inspect.
    output_file (str): The name of the output LaTeX file.
    filter_keyword (str): Include only functions with names containing this keyword.
    """
    # Extract the source code of the module
    source_code = inspect.getsource(module)

    # Use regex to match function definitions in the source code
    functions = re.findall(r"def (\w+)\(", source_code)

    with open(output_file, "w") as f:
        # Write the LaTeX preamble
        f.write(r"""\documentclass{article}
\usepackage[utf8]{inputenc}
\usepackage{hyperref}
\usepackage{listings}
\usepackage{xcolor} % For custom colors
\usepackage{sectsty} % For customizing section fonts
% \subsectionfont{\raggedright} % Allow line breaks and ragged right alignment             
\usepackage{geometry}
% \geometry{margin=1in}
\usepackage{graphicx}
                
\title{TexTOM - Manual}
\author{Moritz Frewein, Moritz Stammer, Marc Allain, Tilman Grünewald}
\begin{document}
\maketitle
\label{toc}
\tableofcontents

% Define a custom style for docstrings
\lstdefinelanguage{docstring}{
    basicstyle=\ttfamily\small, % Monospaced font
    backgroundcolor=\color[HTML]{F5F5F5}, % Light gray background
    frame=single, % Border around the docstring
    rulecolor=\color[HTML]{D6D6D6}, % Border color
    keywordstyle=\color{blue}, % Optional: color for keywords
    breaklines=true, % Wrap long lines
}
                
\subsectionfont{\large\ttfamily\raggedright}

\include{introduction}
\section{Functions}\label{sec:functions}
""")
        
        # f.write("\\section*{Module: " + module.__name__.replace("_", r"\_") + "}\n\n")

        # Loop through functions in the order they appear in the source code
        for name in functions:
            if hasattr(module, name):
                obj = getattr(module, name)
                if inspect.isfunction(obj) and name[0] != "_" and filter_keyword in name:
                    # Escape underscores in the function name
                    escaped_name = name.replace("_", r"\_")
                    escaped_name_2 = name.replace("_", r"")

                    # Get the function signature
                    try:
                        signature = inspect.signature(obj)
                        # Replace underscores in the signature for LaTeX compatibility
                        escaped_signature = str(signature).replace("_", r"\_")
                    except ValueError:
                        escaped_signature = "()"

                    # Get the docstring and escape LaTeX special characters
                    docstring = inspect.getdoc(obj) or "No docstring available."
                    escaped_docstring = docstring#.replace("_", r"\_")

                    # Write the function name, signature, and docstring
                    f.write("\\subsection*{\\texttt{" + escaped_name + escaped_signature + "}}\n")
                    f.write("\\label{fun:" +escaped_name_2+ "}\n")
                    f.write("\\addcontentsline{toc}{subsection}{"+escaped_name+"}\n\n")
                    # f.write("\\label{"+escaped_name+"}\n") # crashes
                    f.write(r"\begin{lstlisting}[language=docstring]" + "\n")
                    f.write(escaped_docstring + "\n")
                    f.write(r"\end{lstlisting}" + "\n\n")
                    f.write(r"\begin{flushright}" + "\n\n")
                    f.write(r"\hyperref[toc]{ToC}" + "\n\n")
                    f.write(r"\end{flushright}" + "\n\n")
                    f.write(r"\input{functions/" + name + "}\n\n")
                    f.write(r"\vspace{5mm}" + "\n\n")
                    f.write(r"\hrule" + "\n\n")
                    # f.write(r"\newpage" + "\n")

                    # create files to be included if necessary
                    with open(f'textom/documentation/functions/{name}.tex','a') as fid:
                        pass        
                    
        # Write the LaTeX end
        f.write(r"\end{document}")
        
        # Write the LaTeX end
        f.write(r"\end{document}")

    print(f"LaTeX document saved to {output_file}")

import textom
generate_latex_doc(textom.textom, 'textom/documentation/textom_documentation.tex')
# run with python textom/documentation/create_documentation.py


import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import matplotlib.colors as colors
from matplotlib.widgets import Slider, Button, RangeSlider, TextBox
from matplotlib.animation import FuncAnimation
import matplotlib.gridspec as gridspec
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from scipy.spatial import ConvexHull
import warnings
import time

from .optimization.fitting import fitting
from .model import model_crystal as cry
from .numba_plugins import masked_average_data
# from 
from ..config import data_type, hex_notation

############################################################################################
############################################################################################

def projection_with_clickable_pixels(data_av, data_indices, data, q, chi, fov, title=''):
    fig,ax = plt.subplots( figsize=(7,7) )
    im = ax.matshow( data_av.T )
    ax.set_title(title)
    cbar = fig.colorbar(im)
    cbar.set_label('Mean intensity', rotation=90)
    fig.tight_layout()

    # Function to handle mouse clicks (printing pixel coordinates)
    def on_click(event):
        if event.inaxes != ax:
            return # does nothing if the click is outside
        if ax.figure.canvas.toolbar.mode == 'zoom rect': return # does nothing when zooming

        x, y = int(np.round(event.xdata)), int(np.round(event.ydata))
        
        idx = np.ravel_multi_index((x,y), fov)
        actual_idx = data_indices[idx]
        fig1,ax1 = plt.subplots( figsize=(6,6) )
        CHi, QQ = np.meshgrid( chi, q )
        pcm = ax1.pcolormesh(QQ, CHi, data[actual_idx].T, shading='auto', cmap='inferno')
        ax1.set_title(f'Image {actual_idx}, x {x}, y{y}')
        ax1.set_xlabel('q / nm^-1')
        ax1.set_ylabel('chi / degree')
        ax1.xaxis.set_label_position('top') 
        ax1.tick_params(top=True, labeltop=True, bottom=False, labelbottom=False)
        cbar1 = fig.colorbar(pcm)
        cbar1.set_label('Intensity', rotation=90)
        fig1.tight_layout()

        # Add a range slider to adjust the color scale
        plt.subplots_adjust(bottom=0.10)#, hspace=0.3)

        ax1_slider = plt.axes([0.25, 0.05, 0.5, 0.03])
        slider1 = RangeSlider(ax1_slider, "Colorscale min/max", 
                            data[actual_idx].min(), data[actual_idx].max(), valinit=(data[actual_idx].min(),data[actual_idx].max()))
        def update1(val):
            ymin, ymax = slider1.val
            pcm.set_clim(ymin, ymax)
            # ax.set_ylim(ymin, ymax)
            fig1.canvas.draw_idle()

        slider1.on_changed(update1)

    # Connect the click event
    fig.canvas.mpl_connect('button_press_event', on_click)

    # Add a range slider to adjust the color scale
    plt.subplots_adjust(bottom=0.10)#, hspace=0.3)

    ax_slider = plt.axes([0.25, 0.05, 0.5, 0.03])
    slider = RangeSlider(ax_slider, "Colorscale min/max", 
                         data_av.min(), data_av.max(), valinit=(data_av.min(),data_av.max()))

    # ax_ymin = plt.axes([0.15, 0.04, 0.3, 0.03])
    # ax_ymax = plt.axes([0.55, 0.04, 0.3, 0.03])
    # slider_ymin = Slider(ax_ymin, 'min', res.min(), res.max(), valinit=res.min())
    # slider_ymax = Slider(ax_ymax, 'max', res.min(), res.max(), valinit=res.max()) 
    def update(val):
        ymin, ymax = slider.val
        im.set_clim(ymin, ymax)
        # ax.set_ylim(ymin, ymax)
        fig.canvas.draw_idle()

    # def update_lim(val):
    #     im.set_clim(slider_ymin.val, slider_ymax.val)
    #     fig.canvas.draw_idle()
    # slider_ymin.on_changed(update_lim)
    # slider_ymax.on_changed(update_lim)
    slider.on_changed(update)

    plt.show()

def loss(loss, title=''):
    f,ax = plt.subplots()
    ax.plot(loss)
    ax.set_xlabel('Iterations')
    ax.set_ylabel('Loss function')
    ax.set_title(title)
    return f,ax

def singlefit2D( fit:fitting, idx, hkl=None,
        polar = False, mode='line',
        title='', log=True, show_off=False,
        output=False, show=False ):
    
    shape = fit.detShape
    PHi, QQ = np.meshgrid( fit.chi, fit.q )
    # PHi = mod.Chi_det.reshape(shape)
    # QQ = mod.Qq_det.reshape(shape)
    
    g,t = fit.gt_mask[ idx ]
    im = fit.image( fit.C, g, t )
    Fit = np.zeros( shape[0]*shape[1], data_type)
    Fit[fit.mask_detector[:,fit.flag_use_peak].flatten()] = im
    FIt = Fit.reshape(shape) #* fit.sim_fac # reshape and scale

    # Data = np.zeros_like(Fit)
    Data = np.full(Fit.shape, np.nan)
    Data[fit.mask_detector[:,fit.flag_use_peak].flatten()] = fit.data[idx]
    DAta = Data.reshape(shape)

    maxval = max( Data.max(), FIt.max() )
    # maxval = FIt.max()

    if mode=='colors':
        if polar:
            f,ax = plt.subplots(1,3,figsize=(8,3),subplot_kw=dict(projection='polar'))
        else:
            f,ax = plt.subplots(1,3,figsize=(8,3),sharey=True)
            PHi = PHi*180/np.pi
        lognrm = colors.LogNorm(vmin=maxval*1e-2, vmax=maxval)
        if log:
            nrm = colors.LogNorm(vmin=maxval*1e-2, vmax=maxval)
            res = abs(DAta-FIt)
            ax[2].set_title('abs(Residuals)')
        else:
            nrm = colors.Normalize(vmin=-maxval, vmax=maxval)
            res = DAta - FIt
            ax[2].set_title('Data-Fit')
        if show_off:
            DAta[DAta<=0] = maxval*1e-2
            FIt[FIt<=0] = maxval*1e-2
        ax[0].pcolormesh(PHi, QQ, DAta, norm=lognrm, cmap='plasma')
        m= ax[1].pcolormesh(PHi, QQ, FIt, norm=lognrm, cmap='plasma')
        m=ax[2].pcolormesh(PHi, QQ, res, norm=nrm, cmap='RdBu_r' )
        ax[0].set_title('Data')
        ax[1].set_title('Fit')
        ax[0].set_xticks([0, 90, 180, 270])
        ax[1].set_xticks([0, 90, 180, 270])
        ax[0].set_xlabel('$\chi$')
        ax[1].set_xlabel('$\chi$')
        ax[0].set_ylabel('q [nm$^{-1}$]')
        f.suptitle(title)
        f.colorbar(m, label='Scattering Intensity', ax=ax[2])
        f.tight_layout()
    
    else: #line
        f = line_qplot(fit.q, fit.chi*180/np.pi, DAta, FIt, hkl, title)

    if output:
        if not os.path.exists('output'):
            os.makedirs('output')
        f.savefig('output/fitimage.pdf')
    
    if show:
        plt.show()

############################################################################################

def line_qplot_old(q, chi, data, fit, hkl=None):
    """Plots data and fit for every q-value above each other
    (in an admittedly weird fashion but it gives a good overview over many
    fits at a time.  )
    The y-scale does not correspond to to the scale of the data, but just
    marks at which q the data is located!

    Parameters
    ----------
    q : ndarray
        radial variable on detector
    chi : ndarray
        azimuthal detector angle
    data : ndarray
        collection of data for each q-value
    fit : ndarray
        collection of fits for each q-value

    Returns
    -------
    _type_
        _description_
    """
    norm=1
    if q.size > 1:
        norm = np.mean(np.abs(np.diff(q))) / max( np.max(data), np.max(fit))
    f, ax = plt.subplots( figsize=(6,9) )
    ax.plot( chi[0], norm*data[0,0]+q[0],'k-x', label='data', linewidth=0.5 )
    ax.plot( chi[0], norm*data[0,0]+q[0],'k:', label='fit' )
    for k in range(q.size):
        mask = data[:,k] > 0
        ax.plot( chi[mask], norm*data[mask, k]+q[k], '-x', linewidth=0.5 )
    # Reset the color cycle
    ax.set_prop_cycle(None)
    for k in range(q.size):
        ax.plot( chi, norm*fit[:,k]+q[k], ':' )
        if hkl:
            plt.text(chi[chi.size//2], q[k]-0.3, hkl[k], va='top', ha='center')

    ax.legend()
    return f, ax

def line_qplot(q, chi, data, fit, hkl=None, title=''):
    """Plots data and fit for every q-value in different axes

    Parameters
    ----------
    q : ndarray
        radial variable on detector
    chi : ndarray
        azimuthal detector angle
    data : ndarray
        collection of data for each q-value
    fit : ndarray
        collection of fits for each q-value

    Returns
    -------
    f : matplotlib figure handle
    """
    f, ax = plt.subplots( q.size, 1, figsize=(6,9), sharex=True )
    ax[0].plot( chi[0], data[0,0],'kx', label='data', linewidth=0.5 )
    ax[0].plot( chi[0], data[0,0],'r', label='fit' )
    for k in range(q.size):
        # mask = data[:,k] > 0
        # ax[k].plot( chi[mask], data[mask, k], 'kx', linewidth=0.5 )
        ax[k].plot( chi, data[:, k], 'kx', linewidth=0.5 )
        ylabel =  f'q = {q[k]:.1f} nm-1\nIntensity'
        if hkl is not None:
            ylabel = str(hkl[k]) + '\n' + ylabel
        ax[k].set_ylabel(ylabel)
    for k in range(q.size):
        ax[k].plot( chi, fit[:,k], 'r' )

    ax[-1].set_xlabel('$\chi$ / deg')
    ax[0].legend()
    f.suptitle(title)
    f.tight_layout()
    return f

############################################################################################

def compare_projection_av( data, fit, title='', log=False, output=False, show=False ):

    maxval = max( data.max(), fit.max() )
    f,ax = plt.subplots(1,3,figsize=(8,3),sharey=True)
    
    ax[0].imshow( data.T, vmax=maxval, cmap='plasma')
    ax[1].imshow( fit.T, vmax=maxval, cmap='plasma')
    ax[2].imshow( np.abs(data-fit).T, vmax=maxval,  cmap='plasma' )
    ax[0].set_title('Data')
    ax[1].set_title('Fit')
    ax[2].set_title('Residuals')
    f.suptitle(title)
    f.tight_layout()

    if output:
        if not os.path.exists('output'):
            os.makedirs('output')
        f.savefig('output/cmp_proj_av.pdf')
    if show:
        plt.show()

############################################################################################

def compare_projection_ori( ori_dat, ori_fit, qs, hkl=None, title='', output=False, show=False ):
    nq = qs.size
    fig = plt.figure(figsize=(10, min(16, 3*nq)))
    gs = gridspec.GridSpec(nq, 3, width_ratios=[1, 1, 0.3])#, wspace=0.4)
    axes = []#np.atleast_2d( ax ) 
    for iq in range(nq):
        if iq==0:
            ax = fig.add_subplot(gs[iq, 0])
        else:
            ax = fig.add_subplot(gs[iq, 0], sharex=axes[0], sharey=axes[0])
        ax.imshow(ori_dat[iq].T, cmap='hsv', vmin=0, vmax=np.pi)   
        ylabel =  f'q = {qs[iq]:.1f} nm-1'
        if hkl is not None:
            ylabel = str(hkl[iq]) + '\n' + ylabel
        ax.set_ylabel( ylabel )
        axes.append(ax)
        ax = fig.add_subplot(gs[iq, 1], sharex=axes[0], sharey=axes[0])
        ax.imshow(ori_fit[iq].T, cmap='hsv', vmin=0, vmax=np.pi)
        axes.append(ax)
        # f.colorbar(m1)
        # f.colorbar(m2)
    axes[0].set_title('Data')
    axes[1].set_title('Fit')
    for k, ax in enumerate(axes):
        if k<len(axes)-2:
            ax.tick_params(axis='x',          # changes apply to the x-axis
                        which='both',      # both major and minor ticks are affected
                        bottom=False,      # ticks along the bottom edge are off
                        labelbottom=False) # labels along the bottom edge are off)
        if k%2:
            ax.tick_params(axis='y',          # changes apply to the x-axis
                        which='both',      # both major and minor ticks are affected
                        left=False,      # ticks along the bottom edge are off
                        labelleft=False) # labels along the bottom edge are off)
    ax_cbar = fig.add_subplot(gs[0, 2], projection='polar') 
    round_colorbar(ax_cbar)
    fig.suptitle(title)
    fig.tight_layout()

    if output:
        if not os.path.exists('output'):
            os.makedirs('output')
        fig.savefig('output/cmp_proj_ori.pdf')
    if show:
        plt.show()
        
def image_residuals(fit:fitting, g, fov, hkl=None, title=''):

    res = fit.projection_residuals(g).reshape(fov)
    mask = fit.scanmask[g].reshape(fov)
    # t_mask = fit.gt_mask[fit.gt_mask[:,0]==g, 1]

    fig,ax = plt.subplots( figsize=(7,7) )
    im = ax.matshow( res.T )
    ax.set_title(title)
    cbar = fig.colorbar(im)
    cbar.set_label('Residuals', rotation=90)
    fig.tight_layout()

    # Function to handle mouse clicks (printing pixel coordinates)
    def on_click(event):
        if event.inaxes != ax:
            return # does nothing if the click is outside
        if ax.figure.canvas.toolbar.mode == 'zoom rect': return # does nothing when zooming

        x, y = int(np.round(event.xdata)), int(np.round(event.ydata))
        
        if mask[x,y]:
            idx = np.ravel_multi_index((x,y), fov)
            idx = np.where( np.logical_and(
                fit.gt_mask[:,0]==g, fit.gt_mask[:,1]==idx ))[0][0]
            singlefit2D( fit, idx, hkl,
                title = f'Projection {g}, pixel ({x}/{y})',
                polar = False, mode='line',
                log=True, show_off=False,
                output=False, show=True )
    # Connect the click event
    fig.canvas.mpl_connect('button_press_event', on_click)

    # Add a range slider to adjust the color scale
    plt.subplots_adjust(bottom=0.10)#, hspace=0.3)

    ax_slider = plt.axes([0.25, 0.05, 0.5, 0.03])
    slider = RangeSlider(ax_slider, "Colorscale min/max", 
                         res.min(), res.max(), valinit=(res.min(),res.max()))
    # ax_ymin = plt.axes([0.15, 0.04, 0.3, 0.03])
    # ax_ymax = plt.axes([0.55, 0.04, 0.3, 0.03])
    # slider_ymin = Slider(ax_ymin, 'min', res.min(), res.max(), valinit=res.min())
    # slider_ymax = Slider(ax_ymax, 'max', res.min(), res.max(), valinit=res.max()) 
    def update(val):
        ymin, ymax = slider.val
        im.set_clim(ymin, ymax)
        # ax.set_ylim(ymin, ymax)
        fig.canvas.draw_idle()
    # def update_lim(val):
    #     im.set_clim(slider_ymin.val, slider_ymax.val)
    #     fig.canvas.draw_idle()
    # slider_ymin.on_changed(update_lim)
    # slider_ymax.on_changed(update_lim)
    slider.on_changed(update)

    plt.show()

############################################################################################

def projection_residuals( fit:fitting, show=False ):
    fit.choose_projections('full') ## doesn't work with less projections yet (else modify fit.data_av)
    # if np.all(fit.ns == np.array([0])):
    #     if not hasattr( fit, 'data_av' ):
    #         fit.prepare_fit_c0()
    #     proj = fit.projections_c0( fit.C )
    #     data = fit.data_0D
    #     res = fit.insert_fov( (proj-data)**2 ).sum(axis=1)
    #     proj = fit.insert_fov( proj )
    # else:
    print('\tCalculating projections')
    proj_infov = np.array( [fit.projection( g, fit.C, info=False ) for g in range(fit.scanmask.shape[0])])
    proj_sumtrans = proj_infov.sum(axis=1) # sum over all translations
    proj_sumtrans_fullim = fit.insert_images(proj_sumtrans)

    data_infov = fit.insert_fov_1d(fit.data)
    data_sumtrans = data_infov.sum(axis=1) # sum over all translations
    data_sumtrans_fullim = fit.insert_images(data_sumtrans)

    residuals =  fit.insert_fov(((proj_infov[fit.scanmask]-fit.data)**2).sum(axis=1)).sum(axis=1)

    '''
    Add a panel with rotation and tilt!
    '''

    f,ax = plt.subplots(2, sharex=True)
    ax[0].plot(proj_sumtrans.sum(axis=1), label='Simulation')
    ax[0].plot(data_sumtrans.sum(axis=1), label='Data')
    ax[0].set_ylabel( 'Integrated intensity' )
    ax[0].legend()
    ax[1].plot(residuals)
    ax[1].set_xlabel( 'Projection index' )
    ax[1].set_ylabel( 'Squared residuals' )
    f.suptitle( 'All peaks' )
    f.tight_layout()

    nq = fit.q.size
    f,ax = plt.subplots(nq, sharex=True, figsize=(6,9))
    ax = np.atleast_1d(ax)
    for iq in range(nq):
        ax[iq].set_title( f'q = {fit.q[iq]:.1f} nm-1, {fit.hkl[iq]}' )
        ax[iq].plot(proj_sumtrans_fullim[:,:,iq].sum(axis=1), label='Simulation')
        ax[iq].plot(data_sumtrans_fullim[:,:,iq].sum(axis=1), label='Data')
        ax[iq].set_ylabel( 'Integrated intensity' )
    ax[0].legend()
    ax[-1].set_xlabel( 'Projection index' )
    f.tight_layout()

    if show:
        plt.show()

############################################################################################

def interactive_tomogram( datasets, names, fctn, borders=None,
            minmax=None, sliceplane='z', cmap='inferno', cut=1, save_vid_path=False,
            title='',  fps=5, show=True ):
    if sliceplane == 'x':
        xlabel,ylabel = 'y','z'
        if borders:
            xlim=range(borders[2],borders[3])
            ylim=range(borders[4],borders[5])
            sllim=borders[0:2]
    elif sliceplane == 'y':
        datasets=datasets.transpose(0,2,1,3)
        xlabel,ylabel = 'x','z'
        if borders:
            xlim=range(borders[0],borders[1])
            ylim=range(borders[4],borders[5])
            sllim=borders[2:4]
    else:
        datasets=datasets.transpose(0,3,1,2)
        xlabel,ylabel = 'x','y'
        if borders:
            xlim=range(borders[0],borders[1])
            ylim=range(borders[2],borders[3])
            sllim=borders[4:6]
    
    num_datasets = datasets.shape[0]
    num_images = datasets.shape[1]
    if not borders:
        sllim = [0,num_images]
    # Initial dataset and image index
    current_dataset_index = 0
    current_image_index =  sllim[0]
    current_stack = datasets[current_dataset_index]

    # Initial display: display the first image of the first dataset
    fig, ax = plt.subplots( figsize=(10,8) )
    plt.subplots_adjust(left=0.25, bottom=0.2)
    ax.set_title(title)
    
    # Achieve an adaptive colorbar with tunable borders
    if not minmax:
        minmax = [[
            # np.nanmin(stack),
            # np.nanmax(stack)
            np.percentile(stack[~np.isnan(stack)], cut),
            np.percentile(stack[~np.isnan(stack)], 100-cut)
            ] for stack in datasets]

    # Display the first image in the initial dataset
    img_display = ax.imshow(current_stack[0].T, cmap=cmap,
                             vmin=minmax[0][0],
                             vmax=minmax[0][1],
                             )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if borders:
        x_ticks = ax.get_xticks()  # Positions of the x-ticks
        y_ticks = ax.get_yticks()  # Positions of the y-ticks
        new_x_labels = [f"{int(tick) + xlim[0]}" for tick in x_ticks if tick.is_integer()]
        new_y_labels = [f"{int(tick) + ylim[0]}" for tick in y_ticks if tick.is_integer()]
        ax.set_xticklabels(new_x_labels)
        ax.set_yticklabels(new_y_labels)
        # ax.set_xticks(ticks=np.arange(len(xlim)), labels=xlim)  # Set custom x-axis labels
        # ax.set_yticks(ticks=np.arange(len(ylim)), labels=ylim)  # Set custom y-axis labels
    cbar = fig.colorbar(img_display)
    cbar.ax.set_autoscale_on(True)
    cbar.set_label(names[0])

    # Create a slider for switching between images within a dataset
    ax_image_slider = plt.axes([0.25, 0.1, 0.65, 0.03], facecolor='lightgoldenrodyellow')
    image_slider = Slider(ax_image_slider, sliceplane, sllim[0], sllim[1], valinit=sllim[0], valstep=1)

    if num_datasets > 1:
        # Create a slider for switching between datasets
        ax_dataset_slider = plt.axes([0.25, 0.05, 0.65, 0.03], facecolor='lightblue')
        dataset_slider = Slider(ax_dataset_slider, 'Dataset', 0, num_datasets - 1, valinit=0, valstep=1)

    def update(val):
        current_image_index = int(image_slider.val)  # Get the image index from the slider
        if num_datasets > 1:
            current_dataset_index = int(dataset_slider.val)  # Get the dataset index from the slider
        else:
            current_dataset_index = 0
        current_stack = datasets[current_dataset_index]  # Update the dataset
        current_image = current_stack[current_image_index - sllim[0]]
        img_display.set_data(current_image.T)  # Update the image data
        img_display.set_clim(vmin=minmax[current_dataset_index][0],
                      vmax=minmax[current_dataset_index][1])
        cbar.update_normal(img_display)
        cbar.set_label(names[current_dataset_index])
        # cbar.set_clim(vmin=current_image.min(),vmax=current_image.max())
        # cbar.draw_all() 
        # tcks = np.linspace(current_image.min(),current_image.max(),endpoint=True)
        # cbar.set_ticks(tcks)
        # cbar.update_normal(img_display)
        fig.canvas.draw_idle()  # Redraw the figure

    # Attach the update functions to the sliders
    image_slider.on_changed(update)
    if num_datasets > 1:
        dataset_slider.on_changed(update)

    # Add a button for playing the video
    ax_button = plt.axes([0.1, 0.095, 0.1, 0.04])
    button = Button(ax_button, 'Play')

    playing = [False]  # Use a mutable object to toggle playing state

    def toggle_play(event):
        playing[0] = not playing[0]  # Toggle state
        if playing[0]:
            button.label.set_text("Pause")
            ani.event_source.start()
        else:
            button.label.set_text("Play")
            ani.event_source.stop()

    button.on_clicked(toggle_play)

    # Automatic playback with animation
    def update_frame(i):
        if playing[0]:
            next_frame = (int(image_slider.val) - sllim[0] + 1) % num_images + sllim[0]
            image_slider.set_val(next_frame)

    ani = FuncAnimation(fig, update_frame, frames=num_images, interval=100)  #  Interval in milliseconds

    if save_vid_path:
        toggle_play(None)
        # Save the animation as an MP4 file
        ani.save(save_vid_path, writer='pillow', fps=fps)

    if show:
        # Function to handle mouse clicks (printing pixel coordinates)
        def on_click(event):
            if event.inaxes != ax: return # does nothing if the click is outside
            if ax.figure.canvas.toolbar.mode == 'zoom rect': return # does nothing when zooming
            current_image_index = int(image_slider.val)
            if sliceplane=='x':
                y, z = int(np.round(event.xdata)+xlim[0]), int(np.round(event.ydata+ylim[0]))
                x = current_image_index
            elif sliceplane=='y':
                x, z = int(np.round(event.xdata+xlim[0])), int(np.round(event.ydata+ylim[0]))
                y = current_image_index
            elif sliceplane=='z':
                x, y = int(np.round(event.xdata+xlim[0])), int(np.round(event.ydata+ylim[0]))
                z = current_image_index
            try:
                fctn(x,y,z)
            except:
                print('Was not able to plot ODF, check if an optimization is loaded!')
            # print(f"Clicked on pixel ({x}, {y}) in dataset {current_dataset_index}, image {current_image_index}")

        # Connect the click event
        fig.canvas.mpl_connect('button_press_event', on_click)

        plt.show()

def tomo_hull( data ):
    # Create a 3D scatter plot
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    # # Scatter plot of data points
    # ax.scatter(data[:, 0], data[:, 1], data[:, 2], c='r', marker='o')

    # Convex Hull for the points
    hull = ConvexHull(data)

    # Plot the convex hull with colored faces
    for simplex in hull.simplices:
        # Create a polygon for each face of the convex hull
        triangle = Poly3DCollection([data[simplex]], color='cyan', alpha=0.4)#, edgecolor='k')
        ax.add_collection3d(triangle)

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.grid(False)
    ax.set_xlim([data[:,0].min(), data[:,0].max()])
    ax.set_ylim([data[:,1].min(), data[:,1].max()])
    ax.set_zlim([data[:,2].min(), data[:,2].max()])
    set_axes_equal_3d(ax)

    # Show plot
    plt.show()

def round_colorbar(ax=None, ang_max=180, cmap='hsv'):
    """
    Creates a half-circle colorbar for visualizing orientations from 0? to 180?.

    Parameters:
    ax (matplotlib axis object): axis where to put the colorbar, if None creates an axis, default=None.
    ang_max (int): maximum angle in degree, optimized for 360 or 180, default=180.
    cmap (str): Colormap to use for the colorbar, default 'hsv'.
    """
    if not ax:
        # Create a polar plot
        fig, ax = plt.subplots(subplot_kw={'projection': 'polar'}, figsize=(6, 3))

    # Generate data for the colorbar
    theta = np.linspace(0, ang_max*np.pi/180, 100) 
    r = np.linspace(0.9, 1.0, 2)  # Radius positions
    theta, r = np.meshgrid(theta, r)
    z = theta  # Map the angle to the colormap

    # Normalize and map the data to the colormap
    norm = colors.Normalize(vmin=0, vmax=ang_max*np.pi/180)
    cmap = plt.get_cmap(cmap)

    # Plot the half-circle colorbar
    cbar = ax.pcolormesh(theta, r, z, cmap=cmap, norm=norm, shading='auto')

    # Customize the polar plot
    ax.set_yticks([])  # Remove radial ticks
    ax.set_xticks(np.linspace(0, ang_max*np.pi/180, 
                              num=5 if ang_max<=180 else 8, 
                              endpoint= False if ang_max>=360 else True ))  # Angular ticks
    ax.set_xticklabels([f'{int(np.degrees(tick))}\u02DA' for tick in ax.get_xticks()])
    ax.set_rgrids([])  # Remove grid lines
    ax.set_theta_zero_location('N')  # Set 0 degrees at the top
    ax.set_theta_direction(-1)  # Clockwise direction

    # Limit the plot to the half-circle
    ax.set_thetamin(0)
    ax.set_thetamax(ang_max)

    plt.show()



def set_axes_equal_3d(ax):
    """Set equal scaling for the 3D plot axes."""
    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    y_range = abs(y_limits[1] - y_limits[0])
    z_range = abs(z_limits[1] - z_limits[0])

    max_range = max(x_range, y_range, z_range)

    x_middle = np.mean(x_limits)
    y_middle = np.mean(y_limits)
    z_middle = np.mean(z_limits)

    ax.set_xlim3d([x_middle - max_range / 2, x_middle + max_range / 2])
    ax.set_ylim3d([y_middle - max_range / 2, y_middle + max_range / 2])
    ax.set_zlim3d([z_middle - max_range / 2, z_middle + max_range / 2])

############################################################################################

def check_simulated_dose( fit:fitting ):
    sim_dose = fit.Beams.sum(axis=2).sum(axis=1)
    plt.figure()
    plt.plot(sim_dose)

############################################################################################

def plot_orientations_slice( vectors_slice, scaling=np.array([1]), title='' ):

    # print(scaling.shape)
    # print(vectors_slice.shape)
    # scaling=np.array([1])
    x = np.arange(vectors_slice.shape[0])
    y = np.arange(vectors_slice.shape[1])
    Y,X = np.meshgrid(y,x)
    vectors_slice_flat = vectors_slice.reshape((vectors_slice.size//3, 3))
    scal = scaling.flatten() / scaling.max()

    # Create a 3D scatter plot
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    ax.quiver(X.flatten(),
              Y.flatten(),
              np.zeros(X.size),
              scal*vectors_slice_flat[:,0],
              scal*vectors_slice_flat[:,1],
              scal*vectors_slice_flat[:,2],
              pivot='middle')
    set_axes_equal_3d(ax)
    ax.set_title(title)

    plt.show()

# def make_slice_coordinates(bounds):
#     # get coordinates in the slice
#     x = np.arange(vectors_slice.shape[0])
#     y = np.arange(vectors_slice.shape[1])
#     X,Y = np.meshgrid(x,y)

def array3D_slider( values, coordinates=None, minmax=None, cut=1, cmap='inferno',xlabel='',ylabel='',collabel='',invert_x=False,invert_y=False ):

    if coordinates:
        X,Y = coordinates[0], coordinates[1]

    # Initial dataset and image index
    current_image_index = 0

    # Initial display: display the first image of the first dataset
    fig, ax = plt.subplots( figsize=(10,8) )
    plt.subplots_adjust(left=0.25, bottom=0.2)

    # Achieve an adaptive colorbar with tunable borders
    if not minmax:
        minmax = [
            np.percentile(values[~np.isnan(values)], cut),
            np.percentile(values[~np.isnan(values)], 100-cut)
            ]

    # Display the first image in the initial dataset
    if coordinates:
        img_display = ax.pcolormesh(X,Y,values[current_image_index], cmap=cmap,
                             vmin=minmax[0],
                             vmax=minmax[1],
                             )
    else:
        img_display = ax.imshow(values[current_image_index].T, cmap=cmap,
                             vmin=minmax[0],
                             vmax=minmax[1],
                             )
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if invert_x:
        ax.invert_xaxis()
    if invert_y:
        ax.invert_yaxis()
    cbar = fig.colorbar(img_display)
    cbar.ax.set_autoscale_on(True)
    cbar.set_label(collabel)

    # Create a slider for switching between images within a dataset
    ax_image_slider = plt.axes([0.25, 0.1, 0.65, 0.03], facecolor='lightgoldenrodyellow')
    image_slider = Slider(ax_image_slider, '', 0, values.shape[0]-1, valinit=current_image_index, valstep=1)

    def update(val):
        current_image_index = int(image_slider.val)  # Get the image index from the slider
        current_image = values[current_image_index]
        if coordinates:
            img_display.update({'array': current_image})
        else:
            img_display.set_data(current_image)  # Update the image data
        cbar.update_normal(img_display)
        fig.canvas.draw_idle()  # Redraw the figure

    # Attach the update functions to the sliders
    image_slider.on_changed(update)

    plt.show()


def plot_points_quaternionspace( Q, odf=None, thresh=0.01, att=0.1 ):

    Q=np.atleast_2d(Q)
    extr = np.max(np.abs(Q[:,1:]))

    if np.any(odf):
        odf /= odf.max()
        Q = Q[odf < thresh]
        odf = odf[odf < thresh]
        # Define a color (e.g. red) with variable alpha
        colors = np.zeros((odf.size, 4))  # RGBA
        colors[:, 2] = 1.0           # Blue channel
        colors[:, 3] = (1-odf)*att         # Alpha channel
    else:
        colors='b'

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    ax.scatter(Q[:, 1], Q[:, 2], Q[:, 3], 
            color=colors, marker='o')
    
    ax.set_xlim([-extr,extr])
    ax.set_ylim([-extr,extr])
    ax.set_zlim([-extr,extr])
    set_axes_equal_3d(ax)

    plt.show()
    
def plot_odf_slices_euler( euler_angles, odf, phi2_sections = [0, 45, 65], tol=5 ):
    # euler_angles: phi1, Phi, phi2
    # Plot ?2 sections (most common representation)
    # phi2_sections = [0, 45, 65] # degrees
    phi1, Phi, phi2 = euler_angles[:,0]*180/np.pi,euler_angles[:,1]*180/np.pi,euler_angles[:,2]*180/np.pi
    # fig, axes = plt.subplots(1, len(phi2_sections), figsize=(15, 5))
    # for ax, val in zip(axes, phi2_sections):
    #     # find closest phi2 index
    #     idx = (np.abs(phi2 - val)).argmin()
    #     section = odf[:, :, idx].T # transpose so Phi is y-axis


    #     im = ax.imshow(
    #     section,
    #     origin='lower',
    #     extent=[phi1[0], phi1[-1], Phi[0], Phi[-1]],
    #     aspect='auto',
    #     cmap='viridis'
    #     )
    #     ax.set_title(f"?? = {phi2[idx]:.1f}??")
    #     ax.set_xlabel("?? (??)")
    #     ax.set_ylabel("? (??)")
    #     fig.colorbar(im, ax=axes.ravel().tolist(), shrink=0.8, label="ODF intensity")
    #     plt.tight_layout()
    #     plt.show()

    fig, axes = plt.subplots(1, len(phi2_sections), figsize=(15, 5))
    for ax, val in zip(axes, phi2_sections):
        mask = np.abs(phi2 - val) < tol
        h = ax.hexbin(phi1[mask], Phi[mask], C=odf, gridsize=50, cmap='viridis')
        ax.set_title(f"?? ? {val}??")
        ax.set_xlabel("?? (??)")
        ax.set_ylabel("? (??)")

    fig.colorbar(h, ax=axes.ravel().tolist(), shrink=0.8, label="Counts")
    plt.tight_layout()
    plt.show()


def plot_odf_slices_rodrigues( r, odf, r3_slices = [0.0, 1.0, -1.0], tol=0.1 ):

    # Choose slices in r3
    r3_slices = [0.0, 1.0, -1.0]

    fig, axes = plt.subplots(1, len(r3_slices), figsize=(15, 5))
    for ax, val in zip(axes, r3_slices):
        mask = np.abs(r[:,2] - val) < tol
        h = ax.hexbin(r[:,0][mask], r[:,1][mask], C=odf, gridsize=50, cmap='viridis')

        # idx = (np.abs(r - val)).argmin()
        # section = odf[:, :, idx].T

        # im = ax.imshow(
        #     section,
        #     origin='lower',
        #     extent=[r[0], r[-1], r[0], r[-1]],
        #     aspect='equal',
        #     cmap='viridis'
        # )
        ax.set_title(f"r? = {val:.2f}")
        ax.set_xlabel("r?")
        ax.set_ylabel("r?")

    fig.colorbar(h, ax=axes.ravel().tolist(), shrink=0.8, label="ODF intensity")
    plt.tight_layout()
    plt.show()

def plot_powder_pattern(cif_path, cutoff_structure_factor=1e-4, max_hkl=4, q_min=0, q_max=60, 
                        wavelength=None, q_dat=None, chi_dat=None, I_dat=None, upgrade_pointgroup=True):
    q, hkl, S, M, symmetry = cry.structure_factor_from_cif(cif_path, cutoff_structure_factor, max_hkl,  
                                             q_min=q_min, q_max=q_max, wavelength=wavelength,
                                             powder=True, upgrade_pointgroup=upgrade_pointgroup)
    
    if wavelength:
        # compute theta in radians
        theta = np.arcsin(wavelength * q / (4*np.pi))
        Lorentz = 1/np.sin(2*theta)
    else:
        Lorentz = 1

    powder_intensities = S * M * Lorentz
    # plt.plot( q_used, powder_intensities, 'x' )
    # print('hkl\tq\tInt\t\tRelative I\tMultiplicity')
    if hex_notation and symmetry[0]=='6':
        print('hkil\t\tq\tRelative I\tMultiplicity')        
    else:
        print('hkl\t\tq\tRelative I\tMultiplicity')
    # plt.figure()
    # plt.plot([q[0],q[0]],[0,0],'r',label='Powder diffraction pattern')
    for (xi, yi, l, m) in zip(q, powder_intensities, hkl, M):
        # plt.plot([xi,xi],[0,yi],'r')
        # plt.text(xi, yi, l, va='bottom', ha='center')
        # print(f'{l}\t{xi:5.3}\t{yi:8.1f}\t{100*yi/powder_intensities.max():.1f}\t{m}')
        if hex_notation and symmetry[0]=='6':
            print(f'[{l[0]:<2} {l[1]:<2} {l[2]:<2} {l[3]:<2}]\t{xi:5.3}\t{100*yi/powder_intensities.max():.1f}\t\t{m}')
        else:
            print(f'[{l[0]:<2} {l[1]:<2} {l[-1]:<2}]\t{xi:5.3}\t{100*yi/powder_intensities.max():.1f}\t\t{m}')
    # if np.any(q_dat):
    #     
    #     q_dat = q_dat[msk_q]
    #     I_dat = I_dat[msk_q] * powder_intensities.max() / I_dat[msk_q].max()
    #     plt.plot(q_dat, I_dat, label='Data')
    #     plt.legend()
    # plt.ylim(bottom=0)
    # plt.xlabel('q / nm^-1')
    # plt.ylabel('Intensity')
    # plt.show(block=True)

    if np.any(q_dat):
        msk_q = np.logical_and(q_dat>q_min, q_dat<q_max)
        compare_data_powder( q_dat[msk_q], chi_dat, I_dat[:,:,msk_q],
                                 q, powder_intensities, hkl )
    else:
        plt.figure()
        for (xi, yi, l) in zip(q, powder_intensities, hkl):
            plt.plot([xi,xi],[0,yi],'r')
            plt.text(xi, yi, l, va='bottom', ha='center')
        plt.ylim(bottom=0)
        plt.xlabel('q / nm^-1')
        plt.ylabel('Intensity')
        # plt.show(block=True)
        plt.show()


def compare_data_powder( q_values, chi_values, data2d_stack, q_powder, I_powder, hkl ):
    # Set up figure and subplots
    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(8, 12), sharex=True,
                                        gridspec_kw={'height_ratios': [2, 1, 1]})

    # Initial plots    
    CHi, QQ = np.meshgrid( chi_values, q_values )
    pcm = ax1.pcolormesh(QQ, CHi, data2d_stack[0].T, shading='auto', cmap='inferno')
    # fig.colorbar(pcm, ax=ax1)
    ax1.set_ylabel('chi [degree]')

    line1, = ax2.plot(q_values, data2d_stack[0].max(axis=0), label='Max intensity')
    line2, = ax2.plot(q_values, data2d_stack[0].mean(axis=0), label='Mean intensity')
    ax2.legend()
    ax2.set_ylabel('Intensity')

    I_powder /= I_powder.max() # normalize to 1
    ax3.plot([q_powder[0],q_powder[0]],[0,0],'r',label='Powder diffraction pattern')
    for (xi, yi, l) in zip(q_powder, I_powder, hkl):
        ax3.plot([xi,xi],[0,yi],'r')
        ax3.text(xi, yi, l, va='bottom', ha='center')
    ax3.set_ylabel('Relative intensity')
    ax3.set_ylim((0,1.1))

    fig.tight_layout()
    plt.subplots_adjust(bottom=0.15)#, hspace=0.3)

    # Slider setup
    ax_idx = plt.axes([0.15, 0.08, 0.7, 0.03])
    ax_yscale = plt.axes([0.25, 0.04, 0.5, 0.03])
    slider_idx = Slider(ax_idx, 'Index', 0, data2d_stack.shape[0] - 1, valinit=0, valstep=1)
    slider_yscale = RangeSlider(ax_yscale, 'Intensity min/max', 
                            data2d_stack.min(), data2d_stack.max(), 
                            valinit=(ax2.get_ylim()[0],ax2.get_ylim()[1]))
    
    global block_slider
    block_slider = False
    def update_idx(val):
        global block_slider
        idx = int(slider_idx.val)
        pcm.set_array(data2d_stack[idx].T)
        line1.set_ydata(data2d_stack[idx].max(axis=0))
        line2.set_ydata(data2d_stack[idx].mean(axis=0))
        if not block_slider:
            slider_yscale.set_val((0,data2d_stack[idx].max()))
            update_lim(0)
        fig.canvas.draw_idle()
    slider_idx.on_changed(update_idx)

    def update_lim(val):
        global block_slider
        ymin, ymax = slider_yscale.val
        ax2.set_ylim(ymin, ymax)
        pcm.set_clim(ymin, ymax)
        block_slider = True
        fig.canvas.draw_idle()
    slider_yscale.on_changed(update_lim)

    plt.show(block=True)


import os
import h5py
import numpy as np 
import hdf5plugin

path = '/dataSATA/data_synchrotron/esrf_id13_2024_05_ch7039_enthesisII/20240501/PROCESSED_DATA/INTEG/cement_pmma/'
# '/dataSATA/data_synchrotron/esrf_id13_2023_09_ihmi1513_biomorphs/ihmi1531/id13/20230914/PROCESSED_DATA/integ/coral_s3/'
# '/dataSATA/data_synchrotron/esrf_id13_2023_09_ihmi1513_biomorphs/ihmi1531/id13/20230914/PROCESSED_DATA/integ/helix_s7'
#'/dataSATA/data_synchrotron/esrf_id13_2024_05_ch7039_enthesisII/20240501/PROCESSED_DATA/INTEG/helix_nuc1/'
#'/dataSATA/data_synchrotron/esrf_id13_2023_09_ihmi1513_biomorphs/ihmi1531/id13/20230914/PROCESSED_DATA/integ/sx_worm/'
# '/dataSATA/data_synchrotron/esrf_id13_2024_2_enthesis/20240215/PROCESSED_DATA/biomorph_blobb/'
#'/dataSATA/data_synchrotron/esrf_id13_2024_2_enthesis/20240215/PROCESSED_DATA/biomorph_leaf/'
file_in = 'analysis/rec_1d/reconstruction_1Dmm_coral_s3_2024_12_03_17h53.h5'
file_out = file_in[:-3]+'_compressed.h5'
model = 'analysis/tomo.h5'

with h5py.File(os.path.join(path,model), 'r') as hf:
    mask = hf['mask_voxels'][()]
    nvox = hf['nVox'][()]
mask_voxels = np.zeros(np.prod(nvox), bool)
mask_voxels[mask] = True 
mask_voxels = mask_voxels.reshape(nvox)

with h5py.File(os.path.join(path,file_in), 'r') as hf_in:
    with h5py.File(os.path.join(path,file_out), 'w') as hf_out:
        hf_out.create_dataset('data', data = hf_in['tomogram'][()], compression=hdf5plugin.LZ4())
        hf_out.create_dataset('mask_voxels', data=mask_voxels, compression=hdf5plugin.LZ4())
        hf_out.create_dataset('radial_units', data=hf_in['radial_units'][()], compression=hdf5plugin.LZ4())


import os, glob, sys, shutil
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
import h5py, hdf5plugin
import matplotlib.pyplot as plt
from importlib import reload # this is for debugging
from time import time

# domestic
from .model import rotation as rot
from .model import symmetries as sym
from .analysis import orix_plugins as orx
from .model.model_textom import model_textom
from .misc import import_module_from_path
from .odf import gridbased as grd
from ..config import data_type

# sample_dir = '/Users/moritz/Documents/papers/benchmark/gen_sample/'

class creator:
    """ A class to that contains the theoretical description of Texture tomography.

    Takes input from a designated input file, see input/input_template.py
    for how it has to be stuctured
    """

    def __init__(self, sample_dir):
        os.makedirs(os.path.join(sample_dir,'analysis'), exist_ok=True)
        os.makedirs(os.path.join(sample_dir,'data_integrated'), exist_ok=True)
        self.sample_dir = sample_dir
        gen_path = os.path.join(sample_dir,'analysis','generation.py')
        self.gen = import_module_from_path('generation', gen_path)

    def setup_generated_sample(self):

        # save alignment file for projectors
        tomogram=self.gen.sample_mask.astype(data_type)
        with h5py.File(os.path.join(self.sample_dir,'analysis','alignment_result.h5'),'w') as hf:
            hf.create_dataset('kappa', data=self.gen.kappa)
            hf.create_dataset('omega', data=self.gen.omega)
            hf.create_dataset('shifts', data=np.zeros((self.gen.kappa.size,2)))
            hf.create_dataset('tomogram', data=tomogram)
            hf.create_dataset('sinogram', data=[])    

        # save voxelmask
        with open(os.path.join(self.sample_dir,'analysis','voxelmask.txt'), 'w') as fid:
            for iv in np.where(self.gen.sample_mask.flatten())[0]:
                fid.write(f'{iv}\n')      
        
        # initialize the model
        self.mod = model_textom( self.sample_dir, classic=False,
                         q_det=self.gen.q_det, # careful! this does nothing
                         chi_det=self.gen.chi_det )
        
    def define_coefficients(self):

        # find out how many voxels and coefficients
        n_voxels = np.prod(self.mod.nVox)
        n_coefficients = self.mod.difflets.shape[1]

        if self.gen.intensity_style == 'uniform':
            intensity = np.ones((n_voxels, n_coefficients), data_type)
        # 
        self.coefficients = np.zeros((n_voxels, n_coefficients), data_type)
        for v in np.atleast_1d(self.mod.mask_voxels):
            if self.gen.coefficient_style == 'single_first':
                self.coefficients[v,0] = 1
            elif self.gen.coefficient_style == 'single_middle':
                self.coefficients[v, self.coefficients.shape[1]//2] = 1
            elif self.gen.coefficient_style == 'single_random':
                idx = np.random.randint( n_coefficients )
                self.coefficients[v, idx] = 1
            else:
                print('\nCoefficient style not recognized\n')
                sys.exit(1)
                
        self.coefficients *= intensity

        with h5py.File(os.path.join(self.sample_dir,'data_integrated','sample_coefficients.h5'), 'w') as hf:
            hf.create_dataset('coefficients', data=self.coefficients)
            hf.create_dataset('odf_mode', data=self.mod.odf_mode )
            hf.create_dataset('odf_parameter', data=self.mod.odf_parameter )
            hf.create_dataset('intensity_style', data=self.gen.intensity_style)
            hf.create_dataset('coefficient_style', data=self.gen.coefficient_style)

    def save_projections(self):
        # gen_path = os.path.join(sample_dir,'analysis','generation.py')
        # gen = import_module_from_path('generation', gen_path)

        # n_basis_functions, q_odf = setup_grid(gen.grid_resolution, mod.symmetry)
        # generators = sym.generators(mod.symmetry)

        # # set up sample coefficients
        # sample_bf_coefficients = np.zeros((np.prod(mod.nVox), n_basis_functions), data_type)
        # flatmask = np.where(mod.mask_voxels.flatten())[0]
        # for v in flatmask:
        #     i_dir = np.random.randint(n_basis_functions)
        #     sample_bf_coefficients[v,i_dir] = 1
        
        # Qs = rot.QfromOTP( mod.Gs )
        # Qc = rot.QfromOTP( mod.Gc )

        # for g in range(mod.Beams.shape[0]):
        #     diff_patterns_g = projection( 
        #         Qc, mod.Isc, Qs, mod.Beams, mod.iBeams, mod.detShape, mod.dV,
        #         sample_bf_coefficients, q_odf, gen.grid_resolution, generators )
        
        # # set up sample coefficients
        # n_basis_functions = mod.difflets.shape[1]
        # sample_bf_coefficients = np.zeros((np.prod(mod.nVox), n_basis_functions), data_type)
        # flatmask = np.where(mod.mask_voxels.flatten())[0]
        # for v in flatmask:
        #     i_dir = np.random.randint(n_basis_functions)
        #     sample_bf_coefficients[v,i_dir] = 1
        n_rot = self.mod.Beams.shape[0]
        path_int = os.path.join(self.sample_dir,'data_integrated')
        print('\tCalculating projections')
        for g in range(n_rot):
            diff_patterns_g = self.mod.projection( g, self.coefficients )
            n_trans = diff_patterns_g.shape[0]
            detector_images = np.zeros( (n_trans, diff_patterns_g.shape[1], self.gen.q_det.size),
                                      data_type )
            for q in range(diff_patterns_g.shape[2]):
                upper = np.argmax(self.mod.Qq_det[0,q] < self.gen.q_det)
                detector_images[:,:,upper-1] = diff_patterns_g[:,:,q]

            with h5py.File(os.path.join(path_int,f'gen_sample_proj{g:03d}_2d.h5'), 'w') as hf:
                hf.create_dataset('cake_integ', data=detector_images)        
                hf.create_dataset('fov', data=self.mod.fov)
                hf.create_dataset('radial_units', data=self.gen.q_det )#self.mod.Qq_det[:,0])
                hf.create_dataset('azimuthal_units', data=self.mod.Chi_det[:,0] * 180/np.pi)
                hf.create_dataset('tilt_angle', data=self.mod.Kappa[g])
                hf.create_dataset('rot_angle', data=self.mod.Omega[g])
                hf.create_dataset('ion', data=np.ones(n_trans, data_type) )
                hf.create_dataset('ty', data=self.mod.ty)
                hf.create_dataset('tz', data=self.mod.tz)

            sys.stdout.write(f'\r\t\t{(g+1):d} / {n_rot:d}')#, t/it: {t_it:.2f}, t left: {((Nrot-g-1)*t_it/60):.1f} min' )
            sys.stdout.flush()
        
        print(f'\n\tsaved to {path_int}')

# # @njit(parallel=True)
# def projection( Qc, Isc, Qs, Beams, iBeams, detShape, dV, 
#                sample_bf_coefficients, q_odf, resolution, gen ):
#     diff_patterns_g = np.empty((Beams.shape[1],detShape[0],detShape[1]), data_type)
#     for t in prange(Beams.shape[1]):
#         # project the coefficients
#         iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
#         c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], sample_bf_coefficients )
#         # get the resulting odf from rotated mu
#         odf_proj = np.zeros( Qc.shape[0], data_type )
#         idcs_basis = np.nonzero(c_proj > 0.01 * c_proj.max())[0]
#         for c in idcs_basis:
#             q_mu = rot.quaternion_multiply(  q_odf[c], Qs[g] )
#             odf_proj += c_proj[c] * gaussian_3d(Qc, q_mu, resolution*np.pi/180, gen, dV )
#         # sparse calculate the projections (only points in odf that are high)
#         diff_pattern = np.zeros(Isc.shape[1], data_type)
#         idcs_odf = np.nonzero(odf_proj> 0.01 * odf_proj.max())[0]
#         for h in idcs_odf:
#             diff_pattern += Isc[h] * odf_proj[h]
#         diff_patterns_g[t] = diff_pattern.reshape((detShape[0],detShape[1]))
#     return diff_patterns_g

def plot_basefunction_texture(resolution, mod, g_center=(0,0,0)):
    
    q_center = rot.QfromOTP(np.atleast_2d(g_center)).flatten()
    Qc = rot.QfromOTP( mod.Gc )
    generators = sym.generators(mod.symmetry)
    q_gen = rot.QfromOTP(generators)
    Q_group = rot.generate_group(q_gen)

    # show the center of the distribution
    orx.plot_points_in_fz(q_center,mod.symmetry)
    plt.title('mean')
    # plot gaussian distribution
    # odf = fisher_SO3(Qc,q_center,10/(resolution*np.pi/180)**2,Q_group)
    odf = grd.gaussian_SO3(Qc, q_center, resolution*np.pi/180, Q_group)

    # orx.odf_cloud_general(mod.Gc,
    #     gaussian_3d(Qc,q_center,resolution*np.pi/180,generators,1),
    #     mod.symmetry,num_samples=1000)
    # plt.title('gauss')
    # plot fisher distribution
    orx.plot_odf(rot.QfromOTP(mod.Gc),
        odf,
        mod.symmetry,
        )
    # plt.title('fisher')
    # plot a polefigure
    orx.plot_pole_figure_from_odf(mod.Gc,odf,
                    mod.symmetry, hkl=(1,0,0) )
    orx.plot_pole_figure_from_odf(mod.Gc,odf,
                    mod.symmetry, hkl=(0,1,0) )
    orx.plot_pole_figure_from_odf(mod.Gc,odf,
                    mod.symmetry, hkl=(0,0,1) )
